import os
import re


def iter_files(start_dir, file_type: str | None =None):
    for name in os.listdir(start_dir):
        new_path = os.path.join(start_dir, name)
        if os.path.isdir(new_path):
            if name in ('.venv', '.git', '.idea', '.vscode'):
               continue
            yield from iter_files(new_path, file_type)
        elif os.path.isfile(new_path):
            if not file_type or name.endswith(file_type):
                yield new_path

def get_import_stat(file_path):
    p1 = re.compile(r'^\s*import\s+(?P<name>[a-zA-Z0-9_.]+).*$')
    p2 = re.compile(r'\s*from\s+(?P<name>[a-zA-Z0-9_.]+)+\s+import\s+[a-zA-Z0-9_.]+.*$')
    data = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.rstrip()
            obj = p1.match(line) or p2.match(line)
            if obj:
                data.append((line, obj.groupdict()))
        return data

def get_current_project_package(start_dir):
    for name in os.listdir(start_dir):
        new_path = os.path.join(start_dir, name)
        if os.path.isdir(new_path):
            if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', name):
                yield name




def get_installer_package():
    import subprocess
    result = subprocess.check_output(['pip', 'list'], text=True)
    data = result.splitlines()[2:]
    return dict(row.split() for row in data)

def create_installer_package_func():
    installer_package = get_installer_package()
    def find(name):
        for pn, pv in installer_package.items():
            if compare_name(pn, name):
                return f'{pn}=={pv}'
        return ""

    return find


get_pk = create_installer_package_func()
def find_package_name_mapper():
    import sys
    from pathlib import Path
    path = Path(sys.executable).parent.parent / 'lib' / 'site-packages'
    if not path.exists():
        raise ValueError('无法识别到解释器路径！')
    mapper = {}
    for name in os.listdir(path):
        file_path = os.path.join(path, name, 'top_level.txt')
        if not os.path.exists(file_path) or not os.path.isfile(file_path):
            continue
        name = name.split('-')[0]
        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.rstrip()
                if not line:
                    continue
                pk = get_pk(name)
                if not pk:
                    raise ValueError(f'not found package: {name}')
                mapper[line] = pk
    return mapper

def compare_name(name1: str, name2: str):
    name1 = name1.replace('-', '_').lower()
    name2 = name2.replace('-', '_').lower()
    return name1 == name2



def main():
    all_names = set()
    current_names = set(get_current_project_package('.'))
    for file in iter_files(os.path.abspath('.'), file_type='.py'):
        for stat, name_dict in get_import_stat(file):
            all_names.add(name_dict['name'])
    third_package_names = set()
    for name in all_names:
        if name.startswith('.'):
            continue
        name = name.split('.')[0]
        if name in current_names:
            continue
        third_package_names.add(name)
    package_name_mapper = find_package_name_mapper()
    result = []
    for third_package_name in third_package_names:
        pk = get_pk(third_package_name) or package_name_mapper.get(third_package_name)
        if not pk:
            continue
        # print(third_package_name, pk)
        result.append(pk)
    print('识别到依赖包\n', '\n'.join(result), sep='')
    return result


if __name__ == '__main__':
    main()



# coding=utf-8

"""
@fileName       :   runner.py
@data           :   2024/8/27
@author         :   jiangmenggui@hosonsoft.com
"""
import datetime
import json
import os.path
import queue
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Optional, cast

from jinja2 import Environment, FileSystemLoader, select_autoescape

from lljz_tools.color import Color
from lljz_tools.console_table import ConsoleTable
from lljz_tools.simple_pref_test._task import find_tasks, TaskGroup, TaskSet, _TaskSetFunction
from lljz_tools.simple_pref_test.core import TaskResultSet, init_result, strf_time


class PrefRunner:

    def __init__(
            self,
            __name: str = '性能测试', /, *,
            virtual_users: Optional[int] = 1,
            user_add_interval: Optional[float] = 0.1,
            pre_task: Optional[float] = None,
            run_seconds: Optional[float] = 10,
            save_directory: Optional[str] = None,
    ):
        assert pre_task is not None or (virtual_users is not None and user_add_interval is not None)
        assert pre_task is None or (isinstance(pre_task, (float, int)) and pre_task > 0)
        assert virtual_users is None or (isinstance(virtual_users, int) and virtual_users > 0)
        assert user_add_interval is None or (isinstance(user_add_interval, (float, int)) and user_add_interval >= 0)
        self.name = __name
        self.save_directory = save_directory or os.path.abspath(f'./pref_test_result/{self.name}')
        self.run_seconds = float(run_seconds) if run_seconds is not None else 10.0
        self.virtual_users = virtual_users
        self.user_add_interval = user_add_interval
        self.pre_task = pre_task
        max_workers = max(10, virtual_users or 0)
        if pre_task:
            max_workers = max(max_workers, int(pre_task))
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self.tasks = find_tasks()
        self.running = True

    def _run_with_thread_num(self):
        """按照固定线程数启动"""

        def run():
            group = TaskGroup(self.tasks)
            while True:
                task_set = group.get()
                for task in task_set.tasks:
                    if time.time() - TaskResultSet.start_time > self.run_seconds:
                        group.teardown_all()
                        return

                    task_name = getattr(task, 'name', task.__name__)
                    task_set.setup(task_name)
                    task()
                    task_set.teardown(task_name)

        if self.virtual_users is not None:
            for _ in range(self.virtual_users):
                self.pool.submit(run)
                if self.user_add_interval is not None:
                    time.sleep(float(self.user_add_interval))

    def _run_with_pre_task(self):
        """按照每秒启动的线程数启动"""
        group_queue: queue.Queue[TaskGroup] = queue.Queue()
        k = 1
        if self.pre_task is not None:
            pre_task = float(self.pre_task)
            while pre_task > 10:
                pre_task = pre_task / 10
                k = k * 10
            interval = 1 / pre_task

            def run():
                nonlocal group_queue
                if group_queue.empty():
                    group = TaskGroup(self.tasks)
                else:
                    group = group_queue.get()
                task_set = group.get()
                for task in task_set.tasks:
                    if time.time() - TaskResultSet.start_time > self.run_seconds:
                        group.teardown_all()
                        return
                    task_name = getattr(task, 'name', task.__name__)
                    task_set.setup(task_name)
                    task()
                    task_set.teardown(task_name)
                group_queue.put(group)

            while time.time() - TaskResultSet.start_time < self.run_seconds and self.running:
                for _ in range(k):
                    self.pool.submit(run)
                time.sleep(interval)

    def run_task(self):
        if self.pre_task:
            self._run_with_pre_task()
        else:
            self._run_with_thread_num()

    def print_start_info(self):
        print(f"开始测试：{Color.green(self.name)}")
        if self.pre_task:
            print(
                f"{Color.yellow('==========测试任务启动参数==========')}\n"
                f"   {Color.thin_magenta('每秒任务数')} : {self.pre_task}\n"
                f"     {Color.thin_magenta('运行时间')} : {self.run_seconds}s\n"
                f"     {Color.thin_magenta('任务总数')} : {len(set(self.tasks))}\n"
                f"     {Color.thin_magenta('启动时间')} : {strf_time(TaskResultSet.start_time)}\n"
                f"{Color.yellow('====================================')}\n"
            )
        else:
            print(
                f"{Color.yellow('==========测试任务启动参数==========')}\n"
                f"   {Color.thin_magenta('并发线程数')} : {self.virtual_users}\n"
                f" {Color.thin_magenta('线程启动间隔')} : {self.user_add_interval}s\n"
                f"     {Color.thin_magenta('运行时间')} : {self.run_seconds}s\n"
                f"     {Color.thin_magenta('任务总数')} : {len(set(self.tasks))}\n"
                f"     {Color.thin_magenta('启动时间')} : {strf_time(TaskResultSet.start_time)}\n"
                f"{Color.yellow('====================================')}\n"
            )

    def start(self):
        TaskResultSet.start_time = time.time()
        self.print_start_info()
        try:
            self.run_task()
            self.pool.shutdown(wait=True)
        finally:
            end_time = TaskResultSet.end_time = time.time()
            table_data, rps, qps, response = init_result()
            table = ConsoleTable(table_data, caption="性能测试结果")
            print(table)
            print(f'\n\n{Color.green("测试完成！")}[完成时间：{Color.thin_cyan(strf_time(end_time))}]')
            file = self.save_to_html(table_data, rps, qps, response)
            print(f'结果保存至：{Color.thin_blue(file)}')

    def save_to_html(self, table, rps, qps, response):
        def get_avg(data):
            return int(sum(data) / len(data) * 1000)

        def init_table_data(table_data):
            return {"header": list(table_data[0].keys()), "rows": [list(row.values()) for row in table_data]}

        def init_data(data, func):
            tps = sorted(data.items())
            keys = set()
            for _, v in tps:
                for k in v:
                    keys.add(k)
            values = {k: [] for k in keys}
            for _, v in tps:
                for k in keys:
                    values[k].append(func(k, v))
            return json.dumps({
                'keys': [k.split(' ')[1] for k, v in tps],
                "values": [{"name": k, "data": v} for k, v in values.items()]
            }, ensure_ascii=False)

        if not os.path.exists(self.save_directory):
            os.makedirs(self.save_directory)
        filename = f'{self.name}_{datetime.datetime.now().strftime("%Y-%m-%d %H_%M_%S")}.html'
        file = os.path.join(self.save_directory, filename)
        source = Path(__file__).parent.parent / 'source'
        env = Environment(
            loader=FileSystemLoader(source),
            autoescape=select_autoescape(['html', 'xml'])
        )
        with open(source / 'echart.js', 'r', encoding='u8') as f:
            echart_js = f.read()
        template = env.get_template('result.html')
        run_arguments = [
            f"并发线程数 : {self.virtual_users}",
            f"线程启动间隔 : {self.user_add_interval}s"
        ]

        context = {
            "echartJs": echart_js,
            "title": self.name,
            'arguments': [
                *run_arguments,
                f"运行时间 : {self.run_seconds}s",
                f"任务总数 : {len(set(self.tasks))}",
                f"启动时间 : {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(TaskResultSet.start_time))}",
                f"结束时间 : {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(TaskResultSet.end_time))}"],
            "table": init_table_data(table),
            "rpsData": init_data(rps, lambda k, v: v.get(k, 0)),
            "qpsData": init_data(qps, lambda k, v: v.get(k, 0)),
            "responseData": init_data(response, lambda k, v: get_avg(v.get(k, [0])))
        }
        with open(file, 'w', encoding='u8') as f:
            f.write(template.render(context))
        return file


if __name__ == '__main__':
    pass



# coding=utf-8

"""
@fileName       :   __init__.py
@data           :   2024/8/27
@author         :   jiangmenggui@hosonsoft.com
"""
from ._task import TaskSet, task, mark_task
from .runner import PrefRunner
from .task_config import TaskConfig
from .utils import FilePath, DataBase

if __name__ == '__main__':
    pass



from typing import Any, TypedDict
import urllib.parse
from http.client import HTTPConnection, HTTPResponse
from json import dumps, loads
from logging import Logger

from requests import Session, Response
from urllib3.filepost import choose_boundary

from .. import decorators, RandomTools, logger
from ..color import Color


class URL(str):
    def __init__(self, __url: str, /):
        assert isinstance(__url, str), f'{type(__url)} is not str!'
        __url = __url.strip()
        if __url:
            assert __url.startswith('http://') or __url.startswith('https://'), 'url必须以http://或https://开头'
        if __url.endswith('/'):
            print(Color.yellow("url请不要带有‘/’结尾"))
            __url = __url[:-1]
        self._url = __url

    def __str__(self):
        return self._url

    def __repr__(self):
        return self._url

    @property
    def url(self):
        return self._url

def _get_from_dict(data: dict, *keys: str, default_value: Any =None):
    if not keys:
        raise KeyError('keys不能为空')
    data = {k.lower(): v for k, v in data.items()}
    for key in map(lambda x: x.lower(), keys):
        if key in data:
            return data[key]
    return default_value

def keep_length(text: str, length: int):
    if len(text) > length:
        return text[:length-3] + '...'
    return text

class HTTPClient(Session):
    
    def __init__(self, base_url: URL | str = '', *, timeout=120, debug: bool | Logger = False):
        super().__init__()
        url = base_url if isinstance(base_url, URL) else URL(base_url)
        self._base_url = url.url
        self._timeout = timeout
        self.debug = bool(debug)
        self._logger = debug if isinstance(debug, Logger) else logger
        self._response: Response | None = None
        self._data: Any = None
        self._key: str = RandomTools.random_string(10)

    @property
    def key(self):
        return self._key
    
    @key.setter
    def key(self, value):
        self._key = str(value) if value is not None else ''


    def login(self, *args, **kwargs):
        raise NotImplementedError('登录接口尚未实现，建议获取到token后，将token等信息添加到self.headers中')

    def get_response(self, status_code=200):
        if self._response is None:
            raise RuntimeError('请先发送请求！')
        try:
            self._data = self._response.json()
        except Exception:
            message = self._response.text
        else:
            message = _get_from_dict(self._data, 'message', 'msg', 'error', 'error_msg', 'error_description', str(self._data))
        assert self._response.status_code == status_code, f'http status: {self._response.status_code}, message: {keep_length(message, 200)}'
        return self._response

    def get_response_data(self, status_code=200, /, **kwargs) -> Any:
        """
        获取响应数据，并校验数据
        """
        response = self.get_response(status_code)
        if not self._data:
            if kwargs:
                logger.warning(f'响应结果不是dict格式，无法校验数据！')
            return response.text
        if not isinstance(self._data, dict):
            logger.warning(f'响应结果不是dict格式，无法校验数据！')
            return self._data
        for key, value in kwargs.items():
            if key not in self._data:
                raise KeyError(f'响应数据缺少{key}字段')
            assert self._data[key] == value, f'{key!r}值错误，预期{value!r}、实际{self._data[key]!r}'
        return _get_from_dict(self._data, 'data', 'result')


    check_response = get_response_data

    def __hiddenKey(self, data, hiddenKey: str):
        if not isinstance(data, dict):
            return data
        return {k: '******' if isinstance(k, str) and k.upper() == hiddenKey.upper() else v for k, v in data.items()}

    def request(
            self,
            method,
            url: str | bytes,
            params=None,
            data=None,
            headers=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=False,
            cert=None,
            json=None,
    ):
        run_url = self._base_url + str(url)
        if self.debug:
            method = str(method.upper())
            method_color_factory = {
                'POST': Color.yellow, 
                'GET': Color.green, 
                'DELETE': Color.red, 
                'PUT': Color.magenta
            }   
            method_color = method_color_factory.get(method, Color.cyan)  
            self._logger.info(f'{method_color(method)} - {run_url}', stacklevel=5)
        if self.debug and params:
            self._logger.info(f'params - {self.__hiddenKey(params, 'password')}', stacklevel=5)
        if self.debug and json:
            try:
                json_data = dumps(self.__hiddenKey(json, 'password'), ensure_ascii=False)
                self._logger.info(f'Payload - {json_data}', stacklevel=5)
            except TypeError as e:
                self._logger.error(f'json dumps error: {e}', stacklevel=5)
        if self.debug and data:
            self._logger.info(f'Form data - {data}', stacklevel=5)

        response = super().request(
            method, run_url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )
        self._response = response
        return response


class FastResponse:

    def __init__(self, response: HTTPResponse):
        self._response = response
        self.status_code = response.status
        self.content = response.read()
        self.text = self.content.decode('utf-8')
        self.headers = response.headers
        self._json = None

    def json(self):
        if not self._json:
            self._json = loads(self.text)
        return self._json

    def __repr__(self):
        return f'<FastResponse [{self.status_code}]>'

    def __bool__(self):
        return self.status_code == 200

    def __str__(self):
        return f'<FastResponse [{self.status_code}]>'

class ConnectionsType(TypedDict):
    connection: HTTPConnection
    status: bool

class FastHTTPClient:

    def __init__(self, base_url: URL | str, *, timeout=120, debug: bool | Logger = False):
        if base_url.startswith('https://'):
            raise ValueError('FastHTTPClient 不支持https协议')
        if not base_url:
            raise ValueError('base_url 不能为空')
        base_url = base_url if isinstance(base_url, URL) else URL(base_url)
        self._base_url = urllib.parse.urlparse(base_url.url)
        self._timeout = timeout
        self.debug = debug
        self._connections: list[ConnectionsType] = []
        self._closed = False
        self.headers = {}
        self.debug = bool(debug)
        self._logger = debug if isinstance(debug, Logger) else logger

    @decorators.thread_lock
    def __get_connection(self) -> tuple[int, HTTPConnection]:
        if self._closed:
            raise ValueError("FastHTTPClient 已关闭")
        for index, connection in enumerate(self._connections):
            if not connection['status']:
                self._connections[index]['status'] = True
                return index, connection['connection']
        else:
            self._connections.append(
                {
                    "connection": HTTPConnection(self._base_url.netloc, timeout=self._timeout),
                    "status": True
                }
                )
            return len(self._connections) - 1, self._connections[-1]['connection']

    def set_token(self, token: str, key='Authorization'):
        if len(token.split(' ')) == 1:
            token = f"Bearer {token}"
        self.headers[key] = token

    @staticmethod
    def _encode_data_to_form_data(data):
        """将dict转换为multipart/form-data"""
        body = b''
        boundary = f'----{choose_boundary()}'
        content_type = f'multipart/form-data; boundary={boundary}'  # noqa
        for key, value in data.items():
            value = "" if value is None else str(value)
            body += f'--{boundary}\r\n'.encode('utf-8')
            body += f'Content-Disposition: form-data; name="{key}"\r\n\r\n'.encode('utf-8')  # noqa
            body += f'{value}\r\n'.encode('utf-8')
        body += f'--{boundary}--'.encode('utf-8')
        return body, content_type

    def request(self, method, url: str, *, params=None, data=None, json=None, headers=None, **kwargs):
        if url.startswith('http://') or url.startswith('https://'):
            raise ValueError('URL不允许重新指定网络地址')
        if data and json:
            raise ValueError('data和json不能同时存在')
        if params:
            url = f'{url}?{urllib.parse.urlencode(params)}'
        if self.debug:
            method = method.upper()
            method_color = {
                'POST': Color.yellow, 'GET': Color.green, 'DELETE': Color.red, 'PUT': Color.magenta
            }.get(method, Color.cyan)
            self._logger.info(
                f'{method_color(method)} - {self._base_url.scheme}://{self._base_url.netloc}{url}', stacklevel=5
            )
        body = None
        if not headers:
            headers = {}
        if json:
            body = dumps(json, ensure_ascii=True)
            if self.debug:
                self._logger.info(f'Payload - {dumps(json, ensure_ascii=False)}', stacklevel=5)
            headers['Content-Type'] = 'application/json;charset=UTF-8'
        if data:
            body, content_type = self._encode_data_to_form_data(data)
            headers['Content-Type'] = content_type
            if self.debug:
                self._logger.info(f'Form data - {data}', stacklevel=5)
        index, conn = self.__get_connection()
        headers.update(self.headers)
        conn.request(method, url, body=body, headers=headers)
        response = FastResponse(conn.getresponse())
        self._connections[index][1] = False
        return response

    def get(self, url, *, params=None, headers=None):
        return self.request('GET', url, params=params, headers=headers)

    def post(self, url, *, data=None, json=None, headers=None):
        return self.request('POST', url, data=data, json=json, headers=headers)

    def put(self, url, *, data=None, headers=None):
        return self.request('PUT', url, data=data, headers=headers)

    def delete(self, url, *, data=None, headers=None):
        return self.request('DELETE', url, data=data, headers=headers)

    def close(self):
        for connection in self._connections:
            connection['connection'].close()
        self._connections.clear()
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


if __name__ == '__main__':
    pass



import sys
import argparse
from .pkg import main as _main


def main():
    parser = argparse.ArgumentParser(
        description="LLJZ Tools - 开发工具集",
        epilog="示例: lljz-tools pipreqs [output_file]"
    )

    subparsers = parser.add_subparsers(dest='command', help='可用命令')

    # pipreqs 命令
    pipreqs_parser = subparsers.add_parser(
        'pipreqs',
        help='生成requirements.txt文件',
        description='生成Python依赖文件')
    pipreqs_parser.add_argument(
        'output_file',
        nargs='?',
        default='requirements.txt',
        help='输出文件名（默认为requirements.txt）')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    if args.command == 'pipreqs':
        with open(args.output_file, 'w', encoding='utf-8') as f:
            f.write('\n'.join(_main()))
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# coding=utf-8
import datetime
import inspect
import threading
import time
import traceback
from functools import wraps
from typing import Callable, Any, Type

from lljz_tools.color import Color


def cache_with_params(*params: str):
    """
    缓存数据，根据传入的参数进行缓存
    :param params:
    :return:
    """

    def decorator(func: Callable):

        __cache = {}

        @wraps(func)
        def wrapper(*args, **kwargs):
            # 获取func的参数定义
            argspec = inspect.getfullargspec(func)
            call_values = dict(zip(argspec.args, args))
            call_values.update(kwargs)
            call_values.pop('self', None)
            call_values.pop('cls', None)
            if not params:
                key = tuple(sorted(call_values.items()))
            else:
                key = tuple(sorted(item for item in call_values.items() if item[0] in params))
            try:
                if key in __cache:
                    return __cache[key]
                __cache[key] = func(*args, **kwargs)
                return __cache[key]
            except TypeError as e:
                if str(e).startswith('unhashable type'):
                    print(Color.yellow(f"[WARNING]传入的值必须是可哈希才能使用缓存，无法缓存参数：{key}"))
                    return func(*args, **kwargs)
                else:
                    raise e

        return wrapper
    return decorator


def time_cache(__time: float, /):
    """
    给缓存数据设置超时时间，
    :param __time: 缓存时间
    :return:
    """
    __cache = {}

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = (func.__name__, *args, *tuple(kwargs.items()))
            if key in __cache and time.time() - __cache[key][0] <= __time:
                return __cache[key][1]
            res = func(*args, **kwargs)
            __cache[key] = (time.time(), res)
            return res

        return wrapper

    return decorator


def timer(__loop=1):
    """
    函数运行计时，重复的运行函数多次，查看函数最终的运行时间
    :param __loop: 重复次数
    :return:
    """

    def outer(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            print(f'function [{Color.yellow(func.__name__)}] start running at {datetime.datetime.now()}')
            start = time.perf_counter()
            res = None
            for _ in range(__loop):
                res = func(*args, **kwargs)
            end = time.perf_counter()
            print(f'function [{Color.yellow(func.__name__)}]      finished at {datetime.datetime.now()}, '
                  f'loop {Color.green(str(__loop))} times, use time {Color.cyan(f"{(end - start) * 1000:.4f}")} ms')
            return res

        return wrapper

    return outer


def catch_exception(
        exceptions: list[Type[Exception]] | Type[Exception] | None = None,
        print_exception: Any | Callable = traceback.print_exception,
        retry_times: int = 0,
        retry_interval: float = 0,  # unit second
):
    """
    自动捕获异常
    :param exceptions: 异常类，默认为Exception
    :param print_exception: 输出异常的方法，默认traceback.print_exception
    :param retry_times: 重试次数，0表示不重试，默认值0
    :param retry_interval: 重试间隔，默认值0，单位秒
    :return:
    """
    if not exceptions:
        exceptions = [Exception]
    elif not isinstance(exceptions, list | tuple | set):
        exceptions = [exceptions]

    def outer(func: Callable):
        @wraps(func)
        def inner(*args, **kwargs):
            for i in range(retry_times + 1):
                try:
                    return func(*args, **kwargs)
                except tuple(exceptions) as e:

                    if i < retry_times:
                        print(
                            f'[{Color.yellow(func.__name__)}] {Color.magenta(e.__class__.__name__)}: '
                            f'{Color.red(str(e))}, {Color.green(f"retry {i + 1}/{retry_times}")}')
                        time.sleep(retry_interval)
                    else:
                        if callable(print_exception):
                            print_exception(e)

        return inner

    return outer


def retry(
        __retry: int = 1,
        /, *,
        interval: float = 0,  # unit second
        exceptions: list[Type[Exception]] | Type[Exception] | None = None,
):
    """
    自动重试（和catch_exception不同的在于最后一次没有成功不会捕获异常）
    :param __retry: 重试次数，0表示不重试，默认值1
    :param interval: 重试间隔，默认值0，单位秒
    :param exceptions: 异常类，默认为Exception，只有捕获到这个异常才会进行重试
    :return:
    """
    if not exceptions:
        exceptions = [Exception]
    elif not isinstance(exceptions, list | tuple | set):
        exceptions = [exceptions]

    def outer(func: Callable):
        @wraps(func)
        def inner(*args, **kwargs):
            for i in range(__retry):
                try:
                    return func(*args, **kwargs)
                except tuple(exceptions) as e:
                    print(f'[{Color.yellow(func.__name__)}] {Color.magenta(e.__class__.__name__)}: '
                          f'{Color.red(str(e))}, {Color.green(f"retry {i + 1}/{__retry}")}')

                    time.sleep(interval)
            return func(*args, **kwargs)

        return inner

    return outer


def debug(func: Callable):
    """
    执行函数时，自动打印函数执行信息
    :param func:
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        def init_arg(arg: Any):
            arg = f'{arg!r}'
            if len(arg) > 50:
                arg = arg[:50] + '...'
            return arg

        args_str = (*map(lambda x: init_arg(x), args), *map(lambda x: f"{x[0]}={init_arg(x[1])}", kwargs.items()))
        _f_info = (f'{Color.yellow(func.__name__)}{Color.blue("(")}'
                   f'{", ".join(map(str, map(Color.green, map(str, args_str))))}{Color.blue(")")}')
        now = datetime.datetime.now()
        t = time.perf_counter()

        res = func(*args, **kwargs)
        print(
            f'function [{Color.yellow(func.__name__)}] execute info: \n'
            f'  execute :  {_f_info[:500]}\n'
            f' start at :  {now}'
        )
        print(
            f'finish at :  {datetime.datetime.now()}\n'
            f'   result :  {str(res)[:500]}\n'
            f' use time :  {Color.cyan(f"{(time.perf_counter() - t) * 1000:.4f}")} ms'
        )
        return res

    return wrapper


def singleton(cls: str | type = ''):
    """
    实现单例模式
    :param cls:
    :return:
    """
    _instance = {}
    lock = threading.Lock()  # noqa
    if isinstance(cls, str):
        def outer(cls_):
            @wraps(cls_)
            def inner(*args, **kwargs):
                with lock:
                    key = kwargs.get(cls, (str(args[0]) if args else None))
                    key = (cls_, key)
                    if key not in _instance:
                        _instance[key] = cls_(*args, **kwargs)
                    return _instance[key]

            return inner

        return outer
    else:

        @wraps(cls)
        def wrapper(*args, **kwargs):
            with lock:
                if cls not in _instance:
                    _instance[cls] = cls(*args, **kwargs)
                return _instance[cls]

        return wrapper


def count_call(func: Callable):
    """
    统计函数被调用的次数
    :param func:
    :return:
    """
    __attr = {'count': 0}

    @wraps(func)
    def wrapper(*args, **kwargs):
        __attr['count'] += 1
        result = func(*args, **kwargs)
        print(f'[{Color.thin_green(str(datetime.datetime.now())[:-3])}]'
              f'[{Color.yellow(func.__name__)}]已被调用了 {Color.green(str(__attr["count"]))} 次')
        return result

    return wrapper


def rate_limited(__max_called_pre_second: float = 1, /, *, auto_wait=False):
    """
    限制函数被调用的频率
    
    :param __max_called_pre_second: 每秒最大调用次数
    :param auto_wait: 是否自动等待，默认值为False
        - True表示自动等待，
        - False表示不等待，超过频率则会抛出异常
    :return:
    """
    if __max_called_pre_second <= 0:
        raise ValueError('__max_called_pre_second must > 0')

    __attr = {'last_called': None, "interval": 1 / __max_called_pre_second}

    def outer(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            now = time.time()
            if not __attr['last_called']:
                __attr['last_called'] = now

            elif now - __attr['last_called'] < __attr['interval']:
                if auto_wait:
                    time.sleep(__attr['interval'] - (now - __attr['last_called']))
                else:
                    raise RuntimeError(
                        f'[{Color.yellow(func.__name__)}]调用过于频繁，'
                        f'每次调用间隔不得低于 {Color.green(f"{__attr["interval"]:.3f}")} 秒！'
                    )
            __attr['last_called'] = time.time()
            return func(*args, **kwargs)

        return wrapper

    return outer


def thread_lock(func: Callable):
    """
    一个装饰器函数，为被装饰的函数提供线程锁，确保该函数在多线程环境下的正确执行。

    参数:
    - func (Callable): 要被装饰，即添加线程锁功能的函数。

    返回值:
    - 返回一个封装了原函数的新函数，该新函数在执行原函数前会获取线程锁。
    """
    lock = threading.Lock()

    @wraps(func)
    def inner(*args, **kwargs):
        nonlocal lock
        # 在执行原函数前，尝试获取线程锁
        with lock:
            # 获取到锁后，执行原函数，并传递所有接收到的参数
            return func(*args, **kwargs)

    return inner


if __name__ == '__main__':
    # @rate_limited(5, auto_wait=False)
    @count_call
    @debug
    def test(*args, **kwargs):
        ...


    from pydantic import BaseModel


    class Cat(BaseModel):
        companyIds: list[int] = [1]
        status: str = '123'
        modelCode: str = 'dsadsf'
        materialCode: str = 'ffffqwfq'
        colorCode: str = 'ggggh'
        shellCode: str = 'hhhhhhhhhhhhhh'
        size: int = 50
        current: int = 1
        
        def __str__(self) -> str:
            return f'{self.__class__.__name__}(companyIds={self.companyIds!r})'
        
        __repr__ = __str__


    for i in range(2):
        test(1, 2, 'cat', Cat(), None, name='dog', age=Cat())



# coding=utf-8

"""
@fileName       :   db_client_2.py
@data           :   2024/5/11
@author         :   jiangmenggui@hosonsoft.com
"""
from logging import Logger
import re
from dataclasses import asdict, dataclass, field
from typing import Any, Iterable
import threading

from lljz_tools.log_manager import LogManager
from lljz_tools.color import Color
from lljz_tools.string_utils import StringUtils


@dataclass
class DatabaseConfig:
    db_type: str
    user: str
    password: str
    host: str
    port: int
    db: str
    qs: dict = field(default_factory=dict)
    cursorclass: Any = None
    charset: str = 'utf8mb4'


@dataclass
class SSHConfig:
    user: str
    password: str
    host: str
    port: int


def parse_db_uri(uri: str) -> DatabaseConfig:
    """
    解析数据库连接字符串
    """
    obj = re.match(
        r'^(?P<db_type>.+)://(?P<user>.+):(?P<password>.+)@(?P<host>.+):(?P<port>\d+)/(?P<db>.+?)(?:\?(?P<qs>.+))?$',
        uri
    )
    if not obj:
        raise ValueError(
            f'{uri!r} 格式错误！期望的格式为：db_type://user:password@host:port/db?qs')
    database_config = obj.groupdict()
    if database_config['qs']:
        qs = {k: v for k, v in (i.split('=', maxsplit=1)
                                for i in database_config.pop('qs').split('&'))}
        database_config.update(qs=qs)
    else:
        database_config.pop('qs')
    database_config['port'] = int(database_config['port'])
    return DatabaseConfig(**database_config)  # type: ignore


sql_logger = LogManager(
    'print_sql', console_level='DEBUG', file_path=None
).get_logger()

ArgsType = list | tuple | None

# 创建一个线程局部存储对象
thread_local = threading.local()


class MySQLClient:
    """
    MySQL数据库连接池，支持ssh隧道连接，

    默认会自动commit，除非强行指定autocommit参数，例如mysql://user:password@host:port/database?autocommit=false
    """

    def __init__(self, uri: str, ssh_uri: str | None = None, show_sql: bool | Logger = True):
        """
        初始化MySQL连接池

        :param uri: MySQL连接字符串，格式：mysql://user:password@host:port/database
            - user: 用户名
            - password: 密码
            - host: 主机地址
            - port: 端口
            - database: 数据库名
        :param ssh_uri: SSH隧道连接字符串，格式：ssh://user:password@host:port
            - user: 用户名
            - password: 密码
            - host: 主机地址
            - port: 端口，端口可以省略，默认22
        :param show_sql: 是否显示SQL语句
        """
        try:
            import pymysql
            import pymysql.cursors
            from pymysql.cursors import DictCursor
            from dbutils.pooled_db import PooledDB
        except ImportError:
            raise ImportError(
                f'package not installed! use "{Color.color('pip install pymysql dbutils', style='i u yellow')}"')

        database_config = parse_db_uri(uri)
        database_config.cursorclass = DictCursor
        if database_config.db_type.upper() != 'MYSQL':
            raise ValueError(f'数据库类型必须为mysql，你输入的是{database_config.db_type!r}')
        self._ssh = None
        if show_sql is True:
            self.show_sql = sql_logger
        elif show_sql is False:
            self.show_sql = None
        else:
            self.show_sql = show_sql

        if ssh_uri:
            try:
                from lljz_tools.client.utils import setup_ssh_tunnel
            except ImportError:
                raise ImportError(
                    f'package not installed! use "{Color.color('pip install sshtunnel', style='i u yellow')}"')

            self._ssh = setup_ssh_tunnel(database_config.host, database_config.port, ssh_uri)
            database_config.host = '127.0.0.1'
            database_config.port = self._ssh.local_bind_port
        database_config_dict = asdict(database_config)
        database_config_dict.pop('db_type')
        qs = database_config_dict.pop('qs')
        if 'autocommit' not in qs:
            qs['autocommit'] = True
        elif qs['autocommit'].lower() == 'false':
            qs['autocommit'] = False
        else:
            qs['autocommit'] = True

        # 创建连接池
        self._pool = PooledDB(
            creator=pymysql,
            # 其他参数根据你的应用需求调整
            mincached=1,
            maxcached=5,
            maxconnections=10,
            blocking=True,  # 如果没有可用连接，调用者将等待而不是抛出异常
            **qs,
            **database_config_dict
        )

    def close(self):
        self._pool.close()
        if self._ssh:
            self._ssh.stop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self) -> "MySQLConnection":
        # 检查当前线程是否已经有连接
        if not hasattr(thread_local, 'conn'):
            setattr(thread_local, 'conn', dict())
        key = id(self)
        if key not in thread_local.conn or not thread_local.conn[key].is_connected():
            thread_local.conn[key] = MySQLConnection(self._pool.connection(), self.show_sql)
        return thread_local.conn[key]

    def select(self, sql: str, args: ArgsType = None, limit: int = 1000) -> list[dict[str, Any]]:
        return self.connect().select(sql, args, limit)

    def select_one(self, sql: str, args: ArgsType = None) -> dict[str, Any]:
        return self.connect().select_one(sql, args)

    def select_one_or_none(self, sql: str, args: ArgsType = None) -> dict[str, Any] | None:
        return self.connect().select_one_or_none(sql, args)

    def select_all(self, sql: str, args: ArgsType = None) -> list[dict[str, Any]]:
        return self.connect().select_all(sql, args)

    def insert(self, sql: str, args: ArgsType = None) -> int:
        return self.connect().insert(sql, args)

    def insert_data(self, table_name: str, data: Iterable[dict]) -> int:
        return self.connect().insert_data(table_name, data)

    def insert_many(self, sql: str, args: list[tuple | list] | tuple[list | tuple]) -> int:
        return self.connect().insert_many(sql, args)

    def update(self, sql: str, args: ArgsType = None) -> int:
        return self.connect().update(sql, args)

    def delete(self, sql: str, args: ArgsType = None) -> int:
        return self.connect().delete(sql, args)

    def commit(self):
        return self.connect().commit()

    def rollback(self):
        return self.connect().rollback()


class MySQLConnection:

    def __init__(self, conn, logger: Logger | None = None):
        self.conn = conn
        self._logger = logger

    def close(self):
        self.conn.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def print_sql(self, sql: str, args: ArgsType = None):
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)

    def _print_sql(self, cursor: Any, sql: str, args: ArgsType = None):
        if not self._logger:
            return
        try:
            self._logger.debug(cursor.mogrify(sql, args), stacklevel=3)
        except TypeError as e:
            self._logger.error(f'[ERROR SQL] sql: {sql}, args: {args}', stacklevel=3)
            self._logger.exception(e, stacklevel=3)
    def select(self, sql: str, args: ArgsType = None, limit: int = 1000) -> list[dict[str, Any]]:
        """
        查询数据，默认查询1000行

        :param sql: 查询语句
        :param args: 查询参数
        :param limit: 查询行数
        :return: 查询结果
        """
        sql = StringUtils.strip_lines(sql)
        sql_type, _ = sql.split(maxsplit=1)
        if sql_type.upper() == 'SELECT':
            sql = f"SELECT * FROM ({sql}) t LIMIT {limit}"
        sql = sql.replace('?', '%s')
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)
            cursor.execute(sql, args)
            return cursor.fetchmany(limit)

    def select_one(self, sql: str, args: ArgsType = None) -> dict[str, Any]:
        data = self.select_one_or_none(sql, args)
        if not data:
            raise ValueError('query result is empty!')
        return data

    def select_one_or_none(self, sql: str, args: ArgsType = None) -> dict[str, Any] | None:
        """
        查询一条数据

        :param sql: 查询语句
        :param args: 查询参数
        :return: 查询结果
        """
        sql = StringUtils.strip_lines(sql)
        sql_type, _ = sql.split(maxsplit=1)
        if sql_type.upper() == 'SELECT':
            sql = f"SELECT * FROM ({sql}) t LIMIT 1"
        sql = sql.replace('?', '%s')

        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)
            cursor.execute(sql, args)
            return cursor.fetchone()

    def select_all(self, sql: str, args: ArgsType = None) -> list[dict[str, Any]]:
        """
        查询所有数据

        :param sql: 查询语句
        :param args: 查询参数
        :return: 查询结果
        """
        sql = StringUtils.strip_lines(sql)
        sql = sql.replace('?', '%s')
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)
            cursor.execute(sql, args)
            return cursor.fetchall()

    def insert(self, sql: str, args: ArgsType = None) -> int:
        """
        插入数据

        :param sql: 插入语句
        :param args: 插入参数
        :return: 插入的数据ID
        """
        sql = StringUtils.strip_lines(sql)
        sql_type, _ = sql.split(maxsplit=1)
        if sql_type.upper() != 'INSERT':
            raise ValueError(f'NOT INSERT SQL: {sql}')
        sql = sql.replace('?', '%s')
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)
            cursor.execute(sql, args)
            return cursor.lastrowid

    def insert_data(self, table_name: str, data: Iterable[dict]) -> int:
        """
        插入数据

        :param table_name: 表名
        :param data: 数据
        :return: 插入行数
        """
        data = tuple(data)
        if not data:
            raise ValueError('data cannot be empty')
        table_name = self._init_table_name(table_name)
        sql = f"INSERT INTO {table_name} ({','.join(map(lambda x: self._init_field_name(x), data[0].keys()))}) VALUES ({','.join(['%s'] * len(data[0]))})"
        values = [tuple(row.values()) for row in data]
        return self.insert_many(sql, values)

    def insert_many(self, sql: str, args: list[tuple | list] | tuple[list | tuple]) -> int:
        """
        批量插入数据

        :param sql: 插入语句
        :param args: 插入参数
        :return: 插入行数
        """
        if not args:
            raise ValueError('待插入的数据不可为空')
        sql = StringUtils.strip_lines(sql)
        sql_type, _ = sql.split(maxsplit=1)
        if sql_type.upper() != 'INSERT':
            raise ValueError(f'NOT INSERT SQL: {sql}')
        sql = sql.replace('?', '%s')
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args[0])
            cursor.executemany(sql, args)
            return cursor.rowcount

    def update(self, sql: str, args: ArgsType = None) -> int:
        """
        更新数据

        :param sql: 更新语句
        :param args: 更新参数
        :return: 更新行数
        """
        sql = StringUtils.strip_lines(sql)
        sql_type, _ = sql.split(maxsplit=1)
        if sql_type.upper() != 'UPDATE':
            raise ValueError(f'NOT UPDATE SQL: {sql}')
        sql = sql.replace('?', '%s')
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)
            cursor.execute(sql, args)
            return cursor.rowcount

    @staticmethod
    def _init_table_name(val: str):

        if not val or not isinstance(val, str):
            raise RuntimeError(f'invalid table name: {val!r}')
        val = val.replace('`', '')
        if not re.match(r"^([a-zA-Z0-9\-_]{1,64}\.)?[a-zA-Z_][a-zA-Z0-9_]{0,63}$", val):
            raise RuntimeError(f'invalid table name: {val!r}')
        return f'`{val.replace('.', '`.`')}`'

    @staticmethod
    def _init_field_name(val: str):
        if not val or not isinstance(val, str):
            raise RuntimeError('invalid field name: {}')
        val = val.replace('`', '')
        if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]{0,63}$", val):
            raise RuntimeError(f'invalid field name: {val!r}')
        return f'`{val}`'

    def update_data(self, table_name: str, data: Iterable[dict]):
        """
        按照update_dict更新指定表的数据

        :param table_name: 待更新的表名
        :param data: 待更新的数据
        :return:
        """
        if not data:
            raise RuntimeError("data cannot be empty!")
        table_name = self._init_table_name(table_name)
        total = 0
        for row in data:
            keys = ','.join(f'{self._init_field_name(k)}=?' for k in row.keys())
            values = list(row.values())
            values.append(row['id'])
            sql = f"UPDATE {table_name} SET {keys} WHERE id = ?"
            total += self.update(sql, values)
        return total

    def delete_data(self, table_name: str, data: Iterable[dict]):
        table_name = self._init_table_name(table_name)
        id_list = [d['id'] for d in data]
        sql = f"DELETE FROM {table_name} WHERE id in ?"
        return self.delete(sql, [id_list])


    def delete(self, sql: str, args: ArgsType = None) -> int:
        """
        删除数据

        :param sql: 删除语句
        :param args: 删除参数
        :return: 删除行数
        """
        sql = StringUtils.strip_lines(sql)
        sql_type, _ = sql.split(maxsplit=1)
        if sql_type.upper() != 'DELETE':
            raise ValueError(f'NOT DELETE SQL: {sql}')
        sql = sql.replace('?', '%s')
        with self.conn.cursor() as cursor:
            self._print_sql(cursor, sql, args)
            cursor.execute(sql, args)
            return cursor.rowcount

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()

    def is_connected(self) -> bool:
        """检查连接是否可用"""
        try:
            self.conn.ping(reconnect=False)  # 检查连接状态
            return True
        except Exception: # noqa
            return False


if __name__ == '__main__':
    pass



# coding=utf-8

"""
@fileName       :   __init__.py
@data           :   2024/11/15
@author         :   jiangmenggui@hosonsoft.com
"""
from . import columns
from .template import ExcelTemplate, ExcelConfig
if __name__ == '__main__':
    pass




import re

from ..color import Color
from ..log_manager import LogManager


ssh_logger = LogManager(
    'sshtunnel', console_level='WARNING', file_path=None
).get_logger()

def parse_ssh_uri(uri: str):
    obj = re.match(
        r'^ssh://(?P<user>.+):(?P<password>.+)@(?P<host>.+?)(?::(?P<port>[0-9]+))?$', uri)
    if not obj:
        raise ValueError(f'Invalid SSH URI format: {uri}')
    ssh_config = obj.groupdict()
    if 'port' not in ssh_config:
        ssh_config['port'] = 22
    ssh_config['port'] = int(ssh_config['port'])
    return ssh_config

def setup_ssh_tunnel(host: str, port: int, ssh_uri: str):
    try:
        from sshtunnel import SSHTunnelForwarder
    except ImportError:
        raise ImportError(f'package not installed! use "{Color.color('pip install sshtunnel', style='i u yellow')}"')
    ssh_config = parse_ssh_uri(ssh_uri)
    ssh = SSHTunnelForwarder(
        ssh_address_or_host=(ssh_config['host'], ssh_config['port']),
        ssh_username=ssh_config['user'],
        ssh_password=ssh_config['password'],
        remote_bind_address=(host, port),
        logger=ssh_logger
    )
    ssh.start()
    return ssh 


import re
import types
from itertools import cycle
from typing import Any, Mapping, Self, Sequence, TypedDict, Union, get_origin, get_args, overload
from collections.abc import Iterable


class ModelConfig(TypedDict, total=False):
    """
    是否允许额外字段，默认True
    设置允许后，会自动将额外字段添加到模型中
    """
    allow_extra: bool

    """
    是否允许缺失字段，默认True
    设置允许后，缺失的字段默认被设置为None
    """
    allow_missing: bool

    """
    是否只读，默认False
    设置只读后，模型中的字段将无法被修改
    """
    readonly: bool

class ModelMetaclass(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        annotations = {}
        defaults = {}
        config = {}
        
        
        # 从基类继承注解和默认值
        for base in bases:
            if hasattr(base, '__annotations__'):
                annotations.update(base.__annotations__)
            if hasattr(base, '__default__'):
                defaults.update(base.__default__)
            if hasattr(base, '__config__'):
                config.update(base.__config__)
        # 更新当前类的注解和默认值
        if '__annotations__' in attrs:
            annotations.update(attrs['__annotations__'])
        if '__default__' in attrs:
            defaults.update(attrs['__default__'])
        if '__config__' in attrs:
            config.update(attrs['__config__'])
        

        remove_keys = []
        
        for key in attrs.keys():
            if key.startswith('__'):
                continue
            if key in annotations:
                defaults[key] = attrs[key]
                remove_keys.append(key)
        
        for key in remove_keys:
            attrs.pop(key)
        
        # 设置类属性
        attrs['__annotations__'] = annotations
        attrs['__annotations__'].pop('__show__', None)
        attrs['__annotations__'].pop('__config__', None)
        attrs['__default__'] = defaults
        attrs['__config__'] = config
        attrs['__config__'].update(kwargs)
        
        return super().__new__(cls, name, bases, attrs)
    
def is_namedtuple(cls):
    """判断一个类是否为namedtuple"""
    return (
        isinstance(cls, tuple.__class__) and  # type: ignore
        hasattr(cls, '_fields')
    )

class Model(dict, metaclass=ModelMetaclass):
    __show__: str | Sequence[str] | None = None
    __config__: ModelConfig = {'allow_extra': True, 'allow_missing': True, 'readonly': False}

    @staticmethod
    def __init_value(cls_, value: Any, readonly: bool = False):
        if readonly:
            M = ReadOnlyModel
        else:
            M = Model
        if value is None:
            return None
        if cls_ is None:
            if isinstance(value, dict):
                return M(value)
            elif isinstance(value, list):
                return [M.__init_value(None, v, readonly) for v in value]
            elif isinstance(value, tuple):
                return tuple(M.__init_value(None, v, readonly) for v in value)
            elif isinstance(value, set):
                return set(M.__init_value(None, v, readonly) for v in value)
            else:
                return value
        if cls_ is dict:
            cls_ = M
        elif cls_ is Model:
            cls_ = M
        try:
            if get_origin(cls_) is Union:
                return M.__init_value(get_args(cls_)[0], value, readonly)
            elif type(cls_) is types.UnionType:
                args = getattr(cls_, '__args__', [])
                return M.__init_value(args[0], value, readonly)
            elif type(cls_) is types.GenericAlias:
                origin = getattr(cls_, '__origin__', list)
                args = getattr(cls_, '__args__', [])
                if origin is dict:
                    return M({M.__init_value(args[0], k, readonly): M.__init_value(args[1], v, readonly) for k, v in value.items()})
                else:
                    return origin((M.__init_value(c, v, readonly) for c, v in zip(cycle(args), value)))
            elif is_namedtuple(cls_):
                return cls_(*value)
            elif cls_ in (list, tuple, set):
                return cls_(value)
            else:
                return cls_(**value)
        except:
            return value

    def __init__(self, __mapper: Mapping[str, Any] | None = None, **kwargs):
        # 直接保存所有字段，包括额外字段
        all_values = {}
        if __mapper:
            kwargs.update(__mapper)
        for key, value in kwargs.items():
            cls = self.__annotations__.get(key, None)
            all_values[key] = self.__init_value(cls, value, readonly=self.__config__.get('readonly', False))
            
        for key, value in self.__default__.items():
            if key not in all_values:
                all_values[key] = value
        if not self.__config__.get('allow_extra', True):
            for key in kwargs.keys():
                if key not in self.__annotations__.keys():
                    raise ValueError(f"{self.__class__.__name__} not allow extra field '{key}'!")
         
        if self.__config__.get('allow_missing', False):
            for key in self.__annotations__.keys():
                if key not in all_values:
                    all_values[key] = None
        super().__init__(all_values)
    
    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """将字典转换为模型对象"""
        if not isinstance(data, dict):
            raise ValueError(f"Expected dict, got {type(data)}")
            
        converted_data = {}
        for key, value in data.items():
            # 递归处理嵌套的数据结构
            if isinstance(value, dict):
                converted_data[key] = cls.from_dict(value)
            elif isinstance(value, list):
                converted_data[key] = cls.from_list(value)
            else:
                converted_data[key] = value
        return cls(**converted_data)
    
    @classmethod
    def from_list(cls, data: list[dict]) -> list[Self]:
        """将列表转换为模型对象列表"""
        if not isinstance(data, list):
            raise ValueError(f"Expected list, got {type(data)}")
            
        return [
            cls.from_dict(item) if isinstance(item, dict) else (cls.from_list(item) if isinstance(item, list) else item) for item in data
        ]  # type: ignore
    
    @classmethod
    @overload
    def from_object(cls, obj: dict) -> Self:
        ...
    @classmethod
    @overload
    def from_object(cls, obj: list[dict]) -> list[Self]:
        ...

    @classmethod
    @overload
    def from_object[T](cls, obj: T) -> T:
        ...
    

    @classmethod
    def from_object[V](cls, obj: dict[str, V] | list[V] | V):
        if isinstance(obj, dict):
            return cls.from_dict({k: cls.from_object(v) for k, v in obj.items()})
        elif isinstance(obj, Iterable):
            return [cls.from_object(item) for item in obj]
        else:
            return obj

    def __getitem__(self, key: Any) -> Any:
        if not self.__config__.get('allow_missing', False):
            return super().__getitem__(key)
        elif key not in self:
            return None
        return super().__getitem__(key)
    
    def __getattr__(self, key: str) -> Any:
        """支持通过属性方式访问字段"""
        try:
            return self.__getitem__(key)
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")
    
    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith('__'):
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)
    
    def __setitem__(self, key: Any, value: Any) -> None:
        if self.__config__.get('readonly', False):
            raise AttributeError(f"'{self.__class__.__name__}' object is readonly!")
        super().__setitem__(key, value)

    def __str__(self):
        show = self.__show__
        if not show:
            if self.__class__.__name__ in ('Model', 'ReadOnlyModel'):
                show = list(self.keys())
            else:
                show = list(self.__annotations__.keys())  # 显示所有字段
        if isinstance(show, str):
            show = re.split(r'[,\s]+', show.strip())
        val = ', '.join(f"{key}={self[key]!r}" for key in show if key in self)
        return f'{self.__class__.__name__}({val})'

    __repr__ = __str__

class ReadOnlyModel(Model, readonly=True):
    pass

if __name__ == "__main__":
    data = [{"id": "12321", "name": "abc"}]
    model = Model.from_list(data)
    print(model)
    print(model[0].id)
    print(model[0].name)
    # print(model.age)
    pass


# coding=utf-8
import logging
import sys
import traceback

from .log_manager import LogManager

print_raw = print

__print = LogManager(
    'print', console_level='INFO', file_path="/python_print",
    fmt='{asctime} | {name} | {module}:{funcName}:{lineno} - {message}'
).get_logger()


def track_print(*args, sep=' ', end='\n', file=None, flush=False):  # noqa
    __print.info(sep.join(map(str, args)), stacklevel=2)


def patch_print():
    __builtins__.print = track_print


def restore_print():
    __builtins__.print = print_raw


patch_print()


def __traceback_exception_print(self, *, file=None, chain=True):
    """Print the result of self.format(chain=chain) to 'file'."""

    if file is None:
        file = sys.stderr
    for line in self.format(chain=chain):
        print_raw(line, file=file, end="")


def __print_list(extracted_list, file=None):
    """Print the list of tuples as returned by extract_tb() or
    extract_stack() as a formatted stack trace to the given file."""
    if file is None:
        file = sys.stderr
    for item in traceback.StackSummary.from_list(extracted_list).format():
        print_raw(item, file=file, end="")


traceback.TracebackException.print = __traceback_exception_print
traceback.print_list = __print_list


def disable_print():
    __logger = LogManager('print').get_logger()
    __logger.setLevel(logging.ERROR)


if __name__ == '__main__':
    def add():
        print('Hello World!')


    # disable_print()
    add()



import re
from typing import Optional

from lljz_tools.color import Color
from ..models import Model


class MongoDBConfig(Model):
    host: str
    port: int
    database: str
    username: Optional[str] = None
    password: Optional[str] = None
    


class MongoClient:
    def __init__(self, uri: str, ssh_uri: Optional[str] = None):
        self.uri = uri
        self._ssh = None
        
        self.config = self.parse_uri(uri)  # 解析 URI 得到配置
        
        # Validate URI
        self.validate_uri()

        if ssh_uri:
            from lljz_tools.client.utils import setup_ssh_tunnel
            
            self._ssh = setup_ssh_tunnel(self.config.host, self.config.port, ssh_uri)
            self.config.host = '127.0.0.1'
            self.config.port = self._ssh.local_bind_port

        self.client, self.db = self.connect_to_mongo()

    @staticmethod
    def parse_uri(uri: str):
        # 解析 URI 并返回配置字典
        pattern = r'mongodb://(?P<username>.+):(?P<password>.+)@(?P<host>.+):(?P<port>\d+)/(?P<database>[^?]+)?'
        match = re.match(pattern, uri)
        if not match:
            raise ValueError(f'Invalid MongoDB URI format: {uri}')
        
        config = match.groupdict()
        config['port'] = int(config['port'])
        return MongoDBConfig(**config)  # type: ignore
    
    def validate_uri(self):
        if not self.uri:
            raise ValueError("MongoDB URI must be specified.")

    def connect_to_mongo(self):
            
        try:
            from pymongo import MongoClient as BaseMongoClient
            
            client = BaseMongoClient(host=self.config.host, port=self.config.port, username=self.config.username, password=self.config.password)
            database = client[self.config.database]  # 获取默认数据库
            return client, database
        except ImportError:
            raise ImportError(f'package not installed! use "{Color.color('pip install pymongo', style='i u yellow')}"')
        except Exception as e:
            raise ConnectionError(f"Failed to connect to MongoDB: {e}")

    def close(self):
        if self.client:
            self.client.close()
        if self._ssh:
            self._ssh.stop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def find_one(self, collection: str, query: dict):
        return self.db[collection].find_one(query)

    def find(self, collection: str, query: dict):
        return list(self.db[collection].find(query))

    def insert_one(self, collection: str, data: dict):
        return self.db[collection].insert_one(data)

    def insert_many(self, collection: str, data: list[dict]):
        return self.db[collection].insert_many(data)

    def update_one(self, collection: str, query: dict, data: dict):
        return self.db[collection].update_one(query, data)

    def update_many(self, collection: str, query: dict, data: dict):
        return self.db[collection].update_many(query, data)

    def delete_one(self, collection: str, query: dict):
        return self.db[collection].delete_one(query)

    def delete_many(self, collection: str, query: dict):
        return self.db[collection].delete_many(query)




# coding=utf-8

"""
@fileName       :   __init__.py
@data           :   2024/2/8
@author         :   jiangmenggui@hosonsoft.com
"""
from .mongo_client import MongoClient
from .mysql_client import MySQLClient
from .http_client import HTTPClient, FastHTTPClient

__all__ = [
    'MongoClient',
    'MySQLClient',
    'HTTPClient',
    'FastHTTPClient',
]
if __name__ == '__main__':
    pass






import enum
import random


class GroupType(enum.Enum):
    """随机字符允许出现的类型"""
    NUMBER = 'integer'  # 数字
    LETTERS = 'letters'  # 字母
    LETTERS_UPPER = 'lettersUpper'  # 大写字母
    LETTERS_LOWER = 'lettersLower'  # 小写字母
    SPECIAL = 'special'  # 特殊符号









class RandomTools:
    @staticmethod
    def random_upper_or_lower(s: str):
        """将字符中的字母随机转换为大写或者小写"""
        return ''.join(random.choice([i.upper(), i.lower()]) for i in s)

    @staticmethod
    def random_string(size: int | list[int] = 15, groups: list[GroupType | str] | None = None):
        """
        :param size: 随机得到的字符长度，如果给出一个数组，则表示每种字符类型的长度
        :param groups: 包含的字符类型：number（数字）、letters（全部字母）lettersUpper（大写字母）、lettersLower（小写字母）、special（特殊字符）
        :return:
        """
        groups = groups or ['integer', 'letters']
        groups = [group.value if isinstance(group, GroupType) else group for group in groups]
        if isinstance(size, list):
            return ''.join(RandomTools.random_string(i, [j]) for i, j in zip(size, groups))
        charList = []
        for g in groups:
            if g == 'integer':
                charList.extend(list('0123456789'))
            elif g == 'letters':
                charList.extend(list(chr(i) for i in range(ord('a'), ord('z') + 1)))
                charList.extend(list(chr(i) for i in range(ord('A'), ord('Z') + 1)))
            elif g == 'lettersUpper':
                charList.extend(list(chr(i) for i in range(ord('A'), ord('Z') + 1)))
            elif g == 'lettersLower':
                charList.extend(list(chr(i) for i in range(ord('a'), ord('z') + 1)))
            elif g == 'special':
                charList.extend(list('`~!@#$%^&*()_+-={}|:"?><,./;\'[]\\'))
        random.shuffle(list(set(charList)))
        result = ''.join(random.choice(charList) for _ in range(size))
        while result.startswith('0'):
            result = RandomTools.random_integer(size)
        return result


    @staticmethod
    def random_string_series(series: int, size=15, groups: list[str | GroupType] | None = None):
        """
        生成一系列的String，字符前缀都相同，后面的序号不同
        random_string_series(3)
        ['3cO5WiACaDVex9M001', '3cO5WiACaDVex9M002', '3cO5WiACaDVex9M003']  # 一个可能的输出

        :param series: 生成的个数,1~100的整数
        :param size: 字符的长度，实际长度为size+序列号
        :param groups: 随机字符允许出现的类型
        :return:
        """
        series = max(1, min(series, 100))
        s = RandomTools.random_string(size, groups)
        size = max(2, len(str(series)))
        return (f'{s}{i:0>{size}}' for i in range(1, series + 1))

    @staticmethod
    def random_integer(size: int | list[int]):
        """生成一个指定长度的随机数字"""
        return RandomTools.random_string(size, groups=[GroupType.NUMBER])


# coding=utf-8

__version__ = '0.3.10'

from .log_manager import LogManager
from .models import Model, ReadOnlyModel
from .random_tools import RandomTools


logger = LogManager('my-tools', console_level='DEBUG').get_logger()


__all__ = [
    'Model', 'ReadOnlyModel',
    'LogManager',
    'logger', 
    'RandomTools'
]
if __name__ == '__main__':
    logger.debug('debug')
    pass



# coding=utf-8

"""
@fileName       :   task_config.py
@data           :   2024/8/27
@author         :   jiangmenggui@hosonsoft.com
"""
import queue
import random
from csv import reader as csv_reader
from threading import Lock
from typing import Any, Generator, Optional, TypeVar

from lljz_tools.client.db_client import MySQLConnectionPool
from lljz_tools.excel import ExcelReader
from lljz_tools.simple_pref_test.utils import FilePath, DataBase

T = TypeVar('T')

class _TaskConfig[T]:

    def __init__(self, data: list[T]):
        self._data = data
        self.lock = Lock()

    def get(self) -> T:
        raise NotImplementedError()


class _TaskRandomConfig[T](_TaskConfig[T]):

    def get(self) -> T:
        with self.lock:
            return random.choice(self._data)


class _TaskLoopConfig[T](_TaskConfig[T]):

    def __init__(self, data: list[T]):
        super().__init__(data)
        self._data = queue.Queue()
        for item in data:
            self._data.put(item)

    def get(self) -> T:
        if self._data.empty():
            raise StopIteration("数据读取完毕")
        item = self._data.get()
        self._data.put(item)
        return item


class _TaskOnceConfig[T](_TaskConfig[T]):

    def __init__(self, data: list[T]):
        super().__init__(data)
        self._data = queue.Queue()
        for item in data:
            self._data.put(item)

    def get(self) -> T:
        if self._data.empty():
            raise StopIteration("数据读取完毕")
        item = self._data.get()
        return item


class TaskConfig[T]:
    """
    任务配置

    model：数据读取模式
        - loop: 循环读取，当数据读取完毕后，从头开始读取
        - once: 一次性读取，当数据读取完毕后，不再读取
        - random: 每次随机取一个值
    """

    def __init__(
        self, 
        *, 
        data: Optional[list[T]] = None, 
        database: Optional[DataBase] = None, 
        csv: Optional[str] = None,
        excel: Optional[str] = None, 
        model: str = 'loop'
    ):
        assert len([i for i in [data, database, csv, excel] if i is not None]) == 1, \
            "data、database、csv、excel参数必须有且仅有一个不为空"
        
        if data is not None:
            _data = data
        elif database is not None:
            _data = self._read_database(database)
        elif csv is not None:
            _data = self._read_csv(csv)
        elif excel is not None:
            _data = self._read_excel(excel)
        else:
            _data = []

        if model == 'random':
            self._config = _TaskRandomConfig(_data) # type: ignore
        elif model == 'once':
            self._config = _TaskOnceConfig(_data) # type: ignore
        else:
            self._config = _TaskLoopConfig(_data) # type: ignore

    @staticmethod
    def _read_database(database: DataBase) -> list[T]:
        with MySQLConnectionPool(database.uri, database.ssh_url or "", show_sql=database.show_sql) as pool:
            with pool.connect() as mysql:
                return mysql.select_all(database.sql)

    @staticmethod
    def _read_csv(file_path: str) -> list[T]:
        data = []
        with open(file_path, mode='r', encoding='utf-8') as file:
            reader = csv_reader(file)
            title = next(reader)
            for row in reader:
                data.append(dict(zip(title, row)))
        return data

    @staticmethod
    def _read_excel(file_path: str) -> Generator:
        reader = ExcelReader(file_path)
        return reader.read()

    def get(self) -> T:
        return self._config.get()


class TaskConfigGroup[T: Any]:
    def __init__(self, data: dict[int | str, list[T]]):
        self.data = {k: TaskConfig[T](data=v) for k, v in data.items()}

    def get(self, key: int | str) -> T:
        return self.data[key].get()


if __name__ == '__main__':
    pass



# coding=utf-8
from typing import Iterable, Any, Optional

from .color import _Color, Color


class CellValue:
    """
    单元格的值
    """

    def __init__(
            self, rawValue: Any, row: int, column: int, width: int, height: int,
            color: str | tuple[int, int, int] = 'black',
            font=None  # PIL.ImageFont.FreeTypeFont
    ):
        """
        初始化单元格
        :param rawValue: 初始的值
        :param row: 所在的行
        :param column: 所在的列
        :param width: 宽度
        :param height: 高度
        :param color: 颜色
        :param font: 字体
        """
        self.rawValue = rawValue
        self.row = row
        self.column = column
        self.width = width
        self.height = height
        self.color = color
        self.font = font
        self.value = rawValue


class TableColumn:
    def __init__(self, width: int | None = None, color: str | tuple[int, int, int] = 'black'):
        self.color = color
        self._autoWidth = width is None  # None则自动计算宽度
        self.width = width if width is not None else 10

    def setWidth(self, width: int):
        if self._autoWidth:
            self.width = max(self.width, width)


class ConsoleTable:

    def __init__(self, data: Iterable[dict[str, Any]], max_width=100, caption='',
                 caption_color=Color.yellow, title_color=Color.thin_cyan, colorize=True):
        def init_value(val):
            if isinstance(val, str | _Color):
                return val
            if val is None:
                return ''
            return str(val)

        self.colorize = colorize
        self.caption = caption
        self.caption_color = caption_color
        self.title_color = title_color
        self.data = [{str(k): init_value(v) for k, v in row.items()} for row in data]
        self.header = list(self.data[0].keys()) if data else []
        self.max_width = max_width
        self.col_width = []
        self._table_str = ""
        self.col_width = self._get_widths()
        self._table_str = self.make_table_str() if self.colorize else self.make_table_str_without_color()

    @staticmethod
    def _get_string_width(val: str | _Color):
        w = 0
        for v in val:
            if u'\u4e00' <= v <= u'\u9fff' or v in '【】（）—…￥！·、？。，《》：；‘“':
                w += 2
            else:
                w += 1
        return w

    def _get_widths(self):
        """获取列宽度，列宽度为整列数据中的最大数据宽度"""

        col_width = [self._get_string_width(key) for key in self.header]
        for row in self.data:
            for i, key in enumerate(self.header):
                value = row.get(key, '')
                width = min(self._get_string_width(value), self.max_width)
                col_width[i] = max(col_width[i], width)
        return col_width

    def _initValueByWidth(self, val: str, width: int):
        newVal = ""
        curWidth = 0
        for v in val:
            w = self._get_string_width(v)
            if curWidth + w > width:
                return newVal[:-3] + '...'
            curWidth += w
            newVal += v
        return newVal

    def make_table_str(self):
        def format_str(val, width):
            val = self._initValueByWidth(val, width)
            length = self._get_string_width(val)
            left = (width - length) // 2
            right = (width - length) - left
            return f'{" " * left}{val}{" " * right}'

        header = ' | '.join(str(self.title_color(format_str(key, w))) for w, key in zip(self.col_width, self.header))
        if self.caption:
            caption = self.caption_color(format_str(self.caption, sum(self.col_width) + (len(self.col_width) - 1) * 3))
            header = caption + '\n' + header
        rows = [' | '.join(format_str(row.get(key, ""), w) for w, key in zip(self.col_width, self.header)) for row in
                self.data]
        return '\n'.join([header, '=' * (sum(self.col_width) + (len(self.col_width) - 1) * 3)] + rows)

    def make_table_str_without_color(self):
        def format_str(val, width):
            length = self._get_string_width(val)
            left = (width - length) // 2
            right = (width - length) - left
            return f'{" " * left}{val}{" " * right}'

        def get_value(row, key):
            val = row.get(key, "")
            if isinstance(val, _Color):
                return val.raw
            return val

        header = ' | '.join(str(format_str(key, w)) for w, key in zip(self.col_width, self.header))
        if self.caption:
            caption = format_str(self.caption, sum(self.col_width) + (len(self.col_width) - 1) * 3)
            header = caption + '\n' + header
        rows = [' | '.join(format_str(get_value(row, key), w) for w, key in zip(self.col_width, self.header)) for row in
                self.data]
        return '\n'.join([header, '=' * (sum(self.col_width) + (len(self.col_width) - 1) * 3)] + rows)

    def __str__(self):
        return self._table_str

    __repr__ = __str__

    def to_image(
            self,
            *,
            odd_row_color='#f8f9fa', even_row_color='white', header_color='#B4DCFF',
            cell_char_size=40,
            font_path: str | None = None, title_font_path: str | None = None,
            font_size=12, title_font_size: Optional[int] = None
    ):
        """
        将表格转换为图片
        :param odd_row_color: 偶数行颜色，默认为 #f8f9fa
        :param even_row_color: 奇数行颜色，默认为 white
        :param header_color: 表头颜色，默认为 #B4DCFF
        :param cell_char_size: 单元格中的单行文本字符数，默认为40，超过这个长度的文本将会被换行显示
        :param font_path: 字体文件路径，一般采用
        :param title_font_path: 表头字体文件路径，一般采用加粗字体
        :param font_size: 字体大小，默认为12
        :param title_font_size: 表头字体大小，默认为font_size + 2
        :return:
        """
        try:
            from PIL import Image, ImageDraw, ImageFont
        except ModuleNotFoundError:
            raise ModuleNotFoundError('请先安装pillow（pip install pillow -U）')

        def split(value, size):
            v = str(value)
            return '\n'.join(('\n'.join(j[i:i + size] for i in range(0, len(j), size))) for j in v.split('\n'))

        data = self.data

        baseImg = Image.new('RGB', (1920, 1080))
        baseDraw = ImageDraw.Draw(baseImg)
        # 初始化字体
        font_path = font_path or 'simhei.ttf'
        title_font_path = title_font_path or font_path or 'msyhbd.ttc'
        font = ImageFont.truetype(font_path, size=font_size)
        bFont = ImageFont.truetype(title_font_path, size=title_font_size or (font_size + 2))

        captionW, captionH = 0, 0
        if self.caption:
            _, _, captionW, captionH = baseDraw.textbbox((0, 0), self.caption, font=bFont)
            captionW, captionH = int(captionW), int(captionH)
        if not data:
            _, _, w, h = baseDraw.textbbox((0, 0), '暂无数据', font=font)
            totalWidth, totalHeight = max(captionW, int(w)) + 20, int(h) + captionH + 15 + 20
            image = Image.new('RGB', (totalWidth, totalHeight), color=(255, 255, 255))
            draw = ImageDraw.Draw(image)
            if self.caption:
                draw.text(((totalWidth - captionW) // 2, 10), self.caption, font=bFont, fill=(0, 0, 0))
                x1, x2, y1, y2 = (totalWidth - captionW) // 2, (
                        totalWidth + captionW) // 2, 12 + captionH, 14 + captionH
                draw.line(((x1, y1), (x2, y1)), fill=(0, 0, 0))
                draw.line(((x1, y2), (x2, y2)), fill=(0, 0, 0))
            draw.text(((totalWidth - int(w)) // 2, 15 + captionH + 10), '暂无数据', font=font, fill='#3E3F41')
            return image
        table: list[list[CellValue]] = [[] for _ in range(len(data) + 1)]
        widths, heights = [10] * len(data[0].keys()), [10] * (len(data) + 1)
        for i, key in enumerate(data[0].keys()):
            val = split(str(key), cell_char_size)
            _, _, w, h = baseDraw.textbbox((0, 0), val, font=bFont)
            table[0].append(CellValue(val, row=0, column=i, width=int(w), height=int(h) + 2, font=bFont))
            heights[0] = max(heights[0], int(h) + 2)
            widths[i] = max(widths[i], int(w))

        for i, row in enumerate(data, start=1):
            for j, key in enumerate(data[0].keys()):
                rawVal = row.get(key)
                color = 'black'
                if rawVal is None:
                    val = ''
                elif isinstance(rawVal, _Color):
                    val = rawVal.raw
                    color = rawVal.hexValue
                else:
                    val = str(rawVal)
                _, _, w, h = baseDraw.textbbox((0, 0), val, font=font)
                table[i].append(
                    CellValue(val, row=i, column=j, width=int(w), height=int(h) + 2, color=color, font=font)
                )
                widths[j] = max(widths[j], int(w))
                heights[i] = max(heights[i], int(h) + 2)

        totalWidth = max(sum(widths) + (len(widths) + 1) * 10, captionW)
        totalHeight = sum(heights) + (len(heights) + 1) * 12 + (captionH + 15 if self.caption else 0)
        img = Image.new('RGB', (totalWidth, totalHeight), color=(255, 255, 255))
        draw = ImageDraw.Draw(img)
        pHeight = 0
        if self.caption:
            draw.text(((totalWidth - captionW) // 2, 10), self.caption, font=bFont, fill=(0, 0, 0))
            x1, x2, y1, y2 = (totalWidth - captionW) // 2, (totalWidth + captionW) // 2, 12 + captionH, 14 + captionH
            draw.line(((x1, y1), (x2, y1)), fill=(0, 0, 0))
            draw.line(((x1, y2), (x2, y2)), fill=(0, 0, 0))
            pHeight = 15 + captionH
        for i, row in enumerate(table, start=1):
            curHeight = pHeight + i * 12 + sum(heights[:i - 1])
            if i > 1:
                draw.rectangle(
                    [(5, curHeight - 5), (totalWidth - 5, curHeight + heights[i - 1] + 5)],
                    fill=odd_row_color if i % 2 == 1 else even_row_color
                )
            else:
                draw.rectangle(
                    [(5, curHeight - 5), (totalWidth - 5, curHeight + heights[i - 1] + 5)],
                    fill=header_color
                )
            for j, val in enumerate(row, start=1):
                curWidth = j * 10 + sum(widths[:j - 1])
                draw.text(
                    (curWidth, curHeight), val.value,
                    font=val.font,
                    fill=val.color)

            draw.line(
                [(5, curHeight + heights[i - 1] + 5), (totalWidth - 5, curHeight + heights[i - 1] + 5)]
                , fill=(0, 0, 0) if i == 1 else (202, 202, 202),
                width=2)
        return img



if __name__ == '__main__':
    table = ConsoleTable(
        [{'server_name': 'intelligent-platform-product', 'status': Color.green('成功'), 'message': '构建成功'}],
        caption='构建结果',
    )
    print(table)



from typing import Callable, Any, Iterable, Iterator, overload


class StringUtils:

    @staticmethod
    def truncate(func: Callable, data: Iterable[Any]):
        """类似于filter方法，不过遇到第一个不符合条件的数据就会退出循环。"""
        for d in data:
            if func(d):
                break
            yield d

    @staticmethod
    def strip_lines(__val: str, __arg = ' ', *args: str):
        def f(x):
            return x not in {__arg, *args}
        value = StringUtils.remove_empty(__val).splitlines()
        min_size = float('inf')
        for row in value:
            min_size = min(len(list(StringUtils.truncate(f, row))), min_size)
        return '\n'.join(v[min_size:] for v in value)

    @staticmethod
    @overload
    def remove_empty(__val: str) -> str: ...

    @staticmethod
    @overload
    def remove_empty(__val: Iterable[str]) -> Iterator[str]: ...

    @staticmethod
    def remove_empty(__val: str | Iterable[str]):
        if isinstance(__val, str):
            __val = filter(lambda x: x.strip(), __val.splitlines())
            return "\n".join(__val)
        return filter(lambda x: x.strip(), __val)

if __name__ == '__main__':
    print(StringUtils.strip_lines("""
    SELECT * FROM om_customer_order
    where id = 1 
    
    """))



from os import PathLike
from typing import Iterable, Optional
from io import BytesIO

from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet

class ExcelWriter:

    def __init__(self, excel_filename: str | PathLike | BytesIO):
        self._file = excel_filename
        self._sheets: dict[str, Worksheet] = {}
        self._excel = self._get_excel()
    
    def _get_excel(self):
        return Workbook(write_only=True)

    @staticmethod
    def __write_sheet(sheet: Worksheet, data: dict | list):
        if isinstance(data, dict):
            sheet.append(tuple(data.values()))
        else:
            sheet.append(data)

    def write(self, data: Iterable[list | dict], /, *, sheet_name: Optional[str] = None) -> Worksheet:
        sheet_name = str(sheet_name) or f'Sheet{len(self._sheets) + 1}'
        if sheet_name not in self._sheets:
            self._sheets[sheet_name] = self._excel.create_sheet(title=sheet_name)
        if not data:
            return self._sheets[sheet_name]

        data = iter(data)
        first = next(data)
        if isinstance(first, dict):
            self._sheets[sheet_name].append(tuple(first.keys()))  # write_title
        self.__write_sheet(self._sheets[sheet_name], first)
        for r in data:
            self.__write_sheet(self._sheets[sheet_name], r)
        return self._sheets[sheet_name]

    def save(self):
        self._excel.save(self._file)

    def close(self):
        self._excel.close()



# coding=utf-8

"""
@fileName       :   compare.py
@data           :   2024/9/10
@author         :   jiangmenggui@hosonsoft.com
"""
from typing import Any


def compare_dict(__a: dict, __b: dict, *ignore_keys: str):
    """
    比较两部字典，判断数据是否一致
    :param __a: 待比较的字典1
    :param __b: 待比较的字典2
    :param ignore_keys: 忽略比较的key
    :return:
    """
    keys = (set(__a.keys()) | set(__b.keys())) - set(ignore_keys)
    return {key: [__a.get(key), __b.get(key)] for key in keys if __a.get(key) != __b.get(key)}


def compare_table(__a: list[dict], __b: list[dict], *keys: str, alais1='数据1', alais2='数据2') -> list[dict[str, Any]]:
    if len(keys) == 0:
        raise TypeError('compare_table() missing 1 required positional argument: \'keys\'')
    a = sorted(__a, key=lambda x: tuple(x[k] for k in keys))
    b = sorted(__b, key=lambda x: tuple(x[k] for k in keys))
    i = j = 0
    result: list[dict[str, Any]] = []
    while i < len(a) and j < len(b):
        k1 = tuple(a[i][k] for k in keys)
        k2 = tuple(b[j][k] for k in keys)
        if k1 < k2:
            result.append({'type': 'missing', 'key': k1, 'value': a[i], 'message': f'{alais2}中缺少数据'})
            i += 1
        elif k1 > k2:
            result.append({'type': 'missing', 'key': k2, 'value': b[j], 'message': f'{alais1}中缺少数据'})
            j += 1
        else:
            dif = compare_dict(a[i], b[j])
            if dif:
                result.append({'type': 'different', 'key': k1, 'value': dif, 'message': '数据不一致'})
            i += 1
            j += 1
    for k_ in range(i, len(a)):
        result.append({'type': 'missing', 'key': tuple(a[k_][k] for k in keys), 'value': a[k_], 'message': f'{alais2}中缺少数据'})
    for k_ in range(j, len(b)):
        result.append({'type': 'missing', 'key': tuple(b[k_][k] for k in keys), 'value': b[k_], 'message': f'{alais1}中缺少数据'})
    return result

if __name__ == '__main__':
    data = compare_table(
        [{'name': 'tom', 'age': 11, 'sex': 'male'}],
        [{'name': 'tom', 'age': 10, 'sex': 'male'}, {'name': 'jim', 'age': 10, 'sex': 'male'}],
        'name',
        alais1='测试环境',
        alais2='生成环境'
    )
    print(*data, sep='\n')
    pass



# coding=utf-8

"""
@fileName       :   template.py
@data           :   2024/11/15
@author         :   jiangmenggui@hosonsoft.com
"""
from typing import NamedTuple

from openpyxl.cell import WriteOnlyCell
from openpyxl.cell.rich_text import TextBlock, CellRichText
from openpyxl.cell.text import InlineFont
from openpyxl.styles import PatternFill, Color
from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet

from lljz_tools.excel import ExcelReader
from .columns import *


class _ExcelTemplateMeta(type):

    def __new__(cls, name, bases, attrs, total=False, variable=True, check_type=False):
        if name in ['ExcelTemplate']:
            return type.__new__(cls, name, bases, attrs)
        __columns__ = {}
        __columns_mapper__ = {}
        __columns_mapper_reverse__ = {}

        for k, v in attrs.items():
            if not isinstance(v, ExcelColumn):
                continue
            __columns__[k] = v
            name = f'*{v.name}' if v.required else v.name
            __columns_mapper__[k] = name
            __columns_mapper_reverse__[name] = k
        attrs['__columns__'] = __columns__
        attrs['__columns_mapper__'] = __columns_mapper__
        attrs['__columns_mapper_reverse__'] = __columns_mapper_reverse__
        return type.__new__(cls, name, bases, attrs)


class ExcelConfig(NamedTuple):
    # sheet页名称，为None则读取默认页
    sheet_name: str = None
    # 唯一的数据列
    unique_columns: list = []
    # 允许空模板
    allow_empty_template: bool = True
    # 忽略空行
    ignore_empty_row: bool = True
    # Excel起始行
    start_row: int = 1
    # Excel最多读取的行
    end_row: int = None


class ExcelTemplate(metaclass=_ExcelTemplateMeta):
    config = ExcelConfig()

    @classmethod
    def _validate_data(cls, data: list[dict]):
        newData = []
        for row in data:
            row = {k: row.get(v) for k, v in cls.__columns_mapper__.items()}
            newRow = {}
            if not cls.config.ignore_empty_row and not any(bool(val) for val in row.values()):
                raise ValueError('数据不能为空')
            for key, value in row.items():
                if key not in cls.__columns__:
                    continue
                column = cls.__columns__[key]
                newRow[key] = column.validate(value)
            newData.append(newRow)
        if cls.config.unique_columns:
            unique_columns = [tuple(row.get(column) for column in cls.config.unique_columns) for row in newData]
            _data = {}
            for rowNo, row in enumerate(unique_columns, start=2):
                if row in _data:
                    raise ValueError(f'与第{_data[row]}行数据重复')
                _data[row] = rowNo
        if not cls.config.allow_empty_template and not newData:
            raise ValueError('模板不能为空')
        return newData

    @classmethod
    def read_excel(cls, file):
        excel = ExcelReader(file)
        data = excel.read(cls.config.sheet_name, min_row=cls.config.start_row, max_row=cls.config.end_row)
        return cls._validate_data(data)

    @staticmethod
    def _write_title(sheet: Worksheet, columns: dict[str, ExcelColumn]):

        title_fill = PatternFill('solid', fgColor='00B0F0')

        required = TextBlock(InlineFont(rFont='等线', sz=12, b=True, color=Color('fa201c')), '*')
        required.name = '*'
        title_font = InlineFont(rFont='等线', sz=12, b=True)

        def get_title_cell(column: ExcelColumn):
            if column.required:
                return CellRichText(
                    required,
                    TextBlock(title_font, column.name),
                )
            value = CellRichText(TextBlock(title_font, column.name))
            cell = WriteOnlyCell(sheet, value=value)  # noqa
            cell.fill = title_fill
            return cell

        row = [get_title_cell(column) for column in columns.values()]
        sheet.append(row)

    @classmethod
    def create_excel_template(cls, file):
        wb = Workbook(write_only=True)
        sheet: Worksheet = wb.create_sheet(title=cls.config.sheet_name or 'sheet1')
        cls._write_title(sheet, cls.__columns__)
        # 设置数据验证
        for index, (k, v) in enumerate(cls.__columns__.items(), start=1):
            dv = v.get_data_validation(columnIndex=index)
            if dv:
                sheet.data_validations.append(dv)
        wb.save(file)


if __name__ == '__main__':
    pass



from os import PathLike
from io import BytesIO
import os

from openpyxl import Workbook, load_workbook
from openpyxl.worksheet.worksheet import Worksheet

from ..excel.reader import ExcelReader
from ..excel.writer import ExcelWriter


class Excel(ExcelReader, ExcelWriter):

    def __init__(self, filename: str | PathLike | BytesIO):
        super().__init__(filename=filename)
        self._sheets = {name: self._excel[name] for name in self._excel.sheetnames}  

    def _get_excel(self):  # noqa
        if isinstance(self._file, str) or isinstance(self._file, PathLike):
            if not os.path.exists(self._file):
                return Workbook()
            else:
                return load_workbook(self._file, data_only=True)
        else:
            return load_workbook(self._file, data_only=True)

    def write(self, data, /, *, sheet_name=None):
        if not sheet_name:
            sheet_name = self._excel.active.title  # type: ignore
        return super().write(data, sheet_name=sheet_name)

    def __getitem__(self, item) -> Worksheet:
        if not self._excel:
            raise ValueError('excel not open')
        return self._excel[item]  # type: ignore



# coding=utf-8

"""
@fileName       :   task.py
@data           :   2024/8/27
@author         :   jiangmenggui@hosonsoft.com
"""
import functools
import inspect
import random
import sys
import time
from types import FunctionType
from typing import Any, Callable, TypeVar

from lljz_tools import logger
from lljz_tools.client.http_client import HTTPClient
from lljz_tools.simple_pref_test.core import TaskResult, TaskResultSet
from lljz_tools.simple_pref_test.task_config import TaskConfig, TaskConfigGroup

T = TypeVar('T', bound=Callable)

def _task_on_function(func: T) -> T:
    task_name = func.__name__

    @functools.wraps(func)
    def inner(*args: Any, **kwargs: Any) -> Any:
        t1 = time.time()
        try:
            res = func(*args, **kwargs)
            t2 = time.time()
            TaskResultSet.put(TaskResult(name=task_name, start=t1, end=t2))
            return res
        except AssertionError as e:
            t2 = time.time()
            TaskResultSet.put(
                TaskResult(name=task_name, start=t1, end=t2, message=f'断言失败：{str(e)}', success=False))
            logger.error(f'{task_name}失败：{str(e)}', stacklevel=2)
        except Exception as e:
            t2 = time.time()
            TaskResultSet.put(TaskResult(
                name=task_name, start=t1, end=t2, message=f'{e.__class__.__name__}: {str(e)}', success=False
            ))
            logger.exception(f'{task_name}错误：{e}', stacklevel=2)

    setattr(inner, 'is_task', True)
    setattr(inner, 'weight', 1)
    setattr(inner, 'name', task_name)
    return inner  # type: ignore

def _task_on_name(name: str = '', weight: int = 1, recode_result: bool = True) -> Callable[[T], T]:
    if not isinstance(weight, int) or weight < 0:
        raise ValueError("任务权重（weight参数）必须为大于0的整数")

    def outer(func: T) -> T:
        task_name = name or func.__name__

        @functools.wraps(func)
        def inner(*args: Any, **kwargs: Any) -> Any:
            t1 = time.time()
            try:
                res = func(*args, **kwargs)
                t2 = time.time()
                if recode_result:
                    TaskResultSet.put(TaskResult(name=task_name, start=t1, end=t2))
                return res
            except AssertionError as e:
                t2 = time.time()
                if recode_result:
                    TaskResultSet.put(
                        TaskResult(name=task_name, start=t1, end=t2, message=f'断言失败：{str(e)}', success=False))
                logger.error(f'{task_name}失败：{str(e)}', stacklevel=2)
            except Exception as e:
                t2 = time.time()
                if recode_result:
                    TaskResultSet.put(TaskResult(
                        name=task_name, start=t1, end=t2, message=f'{e.__class__.__name__}: {str(e)}', success=False
                    ))
                logger.exception(f'{task_name}错误：{e}', stacklevel=2)

        setattr(inner, 'is_task', True)
        setattr(inner, 'weight', weight)
        setattr(inner, 'name', task_name)
        return inner  # type: ignore

    return outer

def task(name: FunctionType | str = '', /, *, weight: int = 1, recode_result: bool = True):
    if isinstance(name, FunctionType):
        return _task_on_function(name)
    elif isinstance(name, str):
        return _task_on_name(name)
    else:
        raise TypeError(f'不支持的参数类型：{type(name)}')

# 只是用来标记task，不会记录最终的执行耗时结果
mark_task = functools.partial(task, recode_result=False)

class TaskSet:
    client: HTTPClient | None = None
    static: bool = True
    weight: int = 1
    task_config: TaskConfig | None = None
    task_config_group: TaskConfigGroup | None = None

    def __init__(self):
        self._task = []

    def setup_class(self):
        """执行所有任务前都会执行setup_class"""
        pass

    def teardown_class(self):
        """执行所有任务后都会执行teardown_class"""
        pass

    def setup(self, task_name: str):
        """
        每个任务执行前都会执行setup

        :param task_name: 任务名，可以根据__task_name判断需要执行的步骤
        """
        pass

    def teardown(self, task_name: str):
        """
        每个任务执行后都会执行teardown

        :param task_name: 任务名，可以根据__task_name判断需要执行的步骤

        """
        pass

    @property
    def tasks(self):
        if self._task:
            return self._task
        for k in self.__dir__():
            obj = getattr(self, k, None)

            if obj is None:
                continue
            if inspect.ismethod(obj) and getattr(obj, 'is_task', None) is True:
                for _ in range(getattr(obj, 'weight', 1)):
                    self._task.append(obj)
        random.shuffle(self._task)
        return self._task

class _TaskSetFunction:

    def __init__(self, task: Callable):
        self.tasks = [task]

    def setup_class(self):
        """执行所有任务前都会执行setup_class"""
        pass

    def teardown_class(self):
        """执行所有任务后都会执行teardown_class"""
        pass

    def setup(self, task_name: str):
        """
        每个任务执行前都会执行setup

        :param task_name: 任务名，可以根据__task_name判断需要执行的步骤
        """
        pass

    def teardown(self, task_name: str):
        """
        每个任务执行后都会执行teardown

        :param task_name: 任务名，可以根据__task_name判断需要执行的步骤
        """
        pass

class TaskGroup:
    """group中会初始化任务示例，初始化的时候会执行setup_class"""

    def __init__(self, task_group: list):
        self.data: list[TaskSet | _TaskSetFunction] = []
        for v in task_group:
            self.add_task(v)

    def add_task(self, __task):
        if inspect.isfunction(__task):
            instance = _TaskSetFunction(__task)
            self.data.extend(instance for _ in range(getattr(__task, 'weight', 1)))
        elif inspect.isclass(__task):
            instance = __task()
            instance.setup_class()
            self.data.extend(instance for _ in range(getattr(__task, 'weight', 1)))
        else:
            raise ValueError('任务必须为函数或者类')

    def get(self) -> TaskSet | _TaskSetFunction:
        return random.choice(self.data)

    def teardown_all(self):
        for t in set(self.data):
            t.teardown_class()

def find_tasks(*modules):
    modules = (*modules, sys.modules['__main__'])
    task_list = []
    for module in modules:
        for k, v in module.__dict__.items():
            if (inspect.isclass(v) and issubclass(v, TaskSet) and v.__dict__.get('static') is not True
                    or inspect.isfunction(v) and getattr(v, 'is_task', False)):
                task_list.append(v)

    return task_list

if __name__ == '__main__':
    pass



# coding=utf-8

"""
@fileName       :   core.py
@data           :   2024/8/27
@author         :   jiangmenggui@hosonsoft.com
"""
import queue
import time
from collections import defaultdict
from typing import NamedTuple, Any, Callable, TypeVar, cast


class TaskResult(NamedTuple):
    name: str
    start: float
    end: float
    message: str = ""
    success: bool = True

    @property
    def use_time(self):
        return self.end - self.start


class TaskResultSet:
    result_data: queue.Queue[TaskResult] = queue.Queue()
    start_time: float = 0.0
    end_time: float = 0.0

    @classmethod
    def put(cls, task_result: TaskResult):
        cls.result_data.put(task_result)


def strf_time(t: float | int) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))


T = TypeVar('T')
CollectionType = TypeVar('CollectionType', int, list)


def get_init_collection(coll_type: Callable[[], CollectionType]) -> dict[str, defaultdict[str, CollectionType]]:
    start = int(TaskResultSet.start_time)
    end = int(TaskResultSet.end_time)
    return {
        strf_time(i): defaultdict(coll_type)
        for i in range(start, end + 1)
    }


def init_result():
    result: dict[str, dict[str, Any]] = {}
    rps = get_init_collection(int)
    qps = get_init_collection(int)
    response = get_init_collection(list)

    while not TaskResultSet.result_data.empty():
        row = TaskResultSet.result_data.get()
        t1 = strf_time(row.start)
        t2 = strf_time(row.end)
        
        if t2 not in rps:
            rps[t2] = defaultdict(int)
        if t1 not in qps:
            qps[t1] = defaultdict(int)
        if t2 not in response:
            response[t2] = defaultdict(list)

        qps[t1][row.name] += 1
        response[t2][row.name].append(row.use_time)
        if row.success:
            rps[t2][row.name + '(成功)'] += 1
        else:
            rps[t2][row.name + '(失败)'] += 1

        if row.name not in result:
            result[row.name] = {
                "NAME": row.name,
                "START_TIME": row.start,
                "END_TIME": row.end,
                "USE_TIME": [row.use_time],
                "SUCCESS": int(bool(row.success))
            }
        else:
            result[row.name]['START_TIME'] = min(result[row.name]['START_TIME'], row.start)
            result[row.name]['END_TIME'] = max(result[row.name]['END_TIME'], row.end)
            result[row.name]['USE_TIME'].append(row.use_time)
            result[row.name]['SUCCESS'] += int(bool(row.success))

    table_data = []
    for value in result.values():
        use_time = sorted(value["USE_TIME"])
        duration = TaskResultSet.end_time - TaskResultSet.start_time
        table_data.append({
            "任务名称": value['NAME'],
            "执行次数": len(use_time),
            "错误次数": len(use_time) - value['SUCCESS'],
            "成功率": f"{value['SUCCESS'] / len(use_time):.4%}",
            "中位数响应": f"{int(use_time[int(len(use_time) * 0.5)] * 1000)}ms",
            "90%响应": f"{int(use_time[int(len(use_time) * 0.9)] * 1000)}ms",
            "95%响应": f"{int(use_time[int(len(use_time) * 0.95)] * 1000)}ms",
            "平均响应": f"{int(sum(value['USE_TIME']) * 1000 / len(value['USE_TIME']))}ms",
            "最小响应": f"{int(use_time[0] * 1000)}ms",
            "最大响应": f"{int(use_time[-1] * 1000)}ms",
            "吞吐量（RPS）": f'{value["SUCCESS"] / duration:.2f}/s'
        })

    return table_data, rps, qps, response


if __name__ == '__main__':
    pass



import csv
from datetime import datetime
from io import BytesIO
from os import PathLike
from typing import Generator, overload, Literal, Iterable

from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet\



def _read_by_range(data: Iterable, min_range: int, max_range: int):
    """
    只读取范围内的数据
    :param data:
    :param min_range: 从1开始，包含
    :param max_range: 包含
    :return:
    """
    for i, val in enumerate(data, start=1):
        if i < min_range:
            continue
        if i > max_range:
            break
        yield val

class CSVReader:
    def __init__(self, filename: str | PathLike):
        self._file = filename
    
    @overload
    def read(self, has_title: Literal[True] = True, min_row: int | None =None, max_row: int | None =None, min_col: int | None =None, max_col: int | None =None) -> Generator[dict[str, str], None, None]:
        ...
    
    @overload
    def read(self, has_title: Literal[False] = False, min_row: int | None =None, max_row: int | None =None, min_col: int | None =None, max_col: int | None =None) -> Generator[list[str], None, None]:
        ...
    
    def read(self, has_title: bool = True, min_row: int | None =None, max_row: int | None =None, min_col: int | None =None, max_col: int | None =None):
        min_row = min_row or 1
        max_row = max_row or 999999999
        min_col = min_col or 1
        max_col = max_col or 999999999
        assert 1 <= min_row <= max_row <= 999999999 and 1 <= min_col <= max_col <= 999999999, \
            "参数有误，行数范围限制为[1, 999999999]"

        with open(self._file) as f:
            reader = (_read_by_range(row, min_col, max_col) for row in _read_by_range(csv.reader(f), min_row, max_row))

            if has_title:
                title = next(reader)

                yield from (dict(zip(title, row)) for row in reader)
            else:
                yield from reader


type ExcelData = str | float | datetime | None

class ExcelReader:

    def __init__(self, filename: str | PathLike | BytesIO):
        self._file = filename
        if isinstance(self._file, str) and self._file.endswith('.csv'):
            raise ValueError('ExcelReader不支持csv文件，请使用CSVReader')
        self._excel = self._get_excel()

    def _get_excel(self):
        return load_workbook(self._file, read_only=True, data_only=True)

    def _get_sheet(self, sheet: int | None | str) -> Worksheet:
        if isinstance(sheet, int):
            return self._excel.worksheets[sheet]
        elif isinstance(sheet, str):
            return self._excel[sheet]
        s = self._excel.active  
        if not s:
            raise ValueError('Excel不包含默认激活的Sheet，请指定一个Sheet名称或者索引')
        return s
    
    @overload
    def read(self, __sheet: int | None | str = None, /, *, has_title: Literal[True] = True, min_row=None, max_row=None,
             min_col=None, max_col=None) -> Generator[dict, None, None]:
        ...
    
    @overload
    def read(self, __sheet: int | None | str = None, /, *, has_title: Literal[False] = False, min_row=None, max_row=None,
             min_col=None, max_col=None) -> Generator[tuple, None, None]:
        ...

    def read(self, __sheet: int | None | str = None, /, *, has_title: bool = True, min_row=None, max_row=None,
             min_col=None, max_col=None):
        """
        读取Excel的Sheet页中的数据

        如果没有指定索引，则范围从A1开始。如果工作表中没有单元格，则返回一个空元组。

        :param __sheet: sheet页的顺序或名字或默认
        :type __sheet: int | None | str

        :param has_title: 是否包含标题，如果包含标题则每行数据都会以字典的形式返回，否则以元组的形式返回
        :type has_title: bool

        :param min_col: 最小列索引(从1开始)
        :type min_col: int

        :param min_row: 最小行索引(从1开始)
        :type min_row: int

        :param max_col: 最大列索引(从1开始)
        :type max_col: int

        :param max_row: 最大行索引(从1开始)
        :type max_row: int

        :rtype: generator
        """
        sheet = self._get_sheet(__sheet)
        iter_rows = sheet.iter_rows(min_row, max_row, min_col, max_col, values_only=True)
        if has_title:
            title = next(iter_rows)
            yield from (dict(zip(title, r)) for r in iter_rows)
        else:
            yield from iter_rows

    @property
    def sheet_names(self):
        return self._excel.sheetnames

    def close(self):
        self._excel.close()



# coding=utf-8

"""
@fileName       :   utils.py
@data           :   2024/8/27
@author         :   jiangmenggui@hosonsoft.com
"""
import os
from typing import NamedTuple

type FilePath = str | os.PathLike


class DataBase(NamedTuple):
    uri: str
    sql: str
    ssh_url: str | None = None
    show_sql: bool = False
if __name__ == '__main__':
    pass



"""
@file   : color.py
@author : jiangmenggui@hosonsoft.com
@data   : 2024/4/2
"""
import re
from typing import Literal


class _Color:

    def __init__(self, s: str, start, end, name, hexValue):
        self.raw = s
        self.s = start + s + end
        self.start = start
        self.end = end
        self.name = name
        self.hexValue = hexValue

    def __len__(self):
        return self.raw.__len__()

    def __str__(self):
        return self.s.__str__()

    def __repr__(self):
        return self.s.__repr__()

    def __iter__(self):
        return self.raw.__iter__()

    def __add__(self, other):
        return self.s + other

    def __radd__(self, other):
        return other + self.s

    def __format__(self, format_spec):
        return self.start + self.raw.__format__(format_spec) + self.end


class _ColorFactory:

    def __init__(self, name, start, end='\033[0m', *, hexValue: str = ''):
        self.name = name
        self.start = start
        self.end = end
        self._hexValue = hexValue

    def __call__(self, __val: str, /, ):
        return _Color(__val, self.start, self.end, self.name, self._hexValue)

type ColorString = Literal[
    'red', 'green', 'yellow', 'blue', 'purple', 'cyan', 'white',
    'thin_red', 'thin_green', 'thin_yellow', 'thin_blue', 'thin_purple', 'thin_cyan', 'thin_white',
    ]

class Color:
    mapper = {"red": "# FF4050",
              "green": "# 4FC414",
              "yellow": "# E5BF00",
              "cyan": "# 00E5E5",
              "blue": "# 1FB0FF",
              "magenta": "# ED7EED",
              "white": "# FFFFFF", }
    UNDERLINE = '\033[4m'
    red = _ColorFactory('red', '\033[91m', hexValue='#FF4050')
    green = _ColorFactory('green', '\033[92m', hexValue='#4FC414')
    yellow = _ColorFactory('yellow', '\033[93m', hexValue='#E5BF00')
    blue = _ColorFactory('blue', '\033[94m', hexValue='#1FB0FF')
    magenta = _ColorFactory('purple', '\033[95m', hexValue='#ED7EED')
    cyan = _ColorFactory('cyan', '\033[96m', hexValue='#00E5E5')
    white = _ColorFactory('white', '\033[97m', hexValue='#FFFFFF')
    thin_red = _ColorFactory('red', '\033[31m', hexValue='#F0524F')
    thin_green = _ColorFactory('green', '\033[32m', hexValue='#5C962C')
    thin_yellow = _ColorFactory('yellow', '\033[33m', hexValue='#A68A0D')
    thin_blue = _ColorFactory('blue', '\033[34m', hexValue='#3993D4')
    thin_magenta = _ColorFactory('purple', '\033[35m', hexValue='#A771BF')
    thin_cyan = _ColorFactory('cyan', '\033[36m', hexValue='#00A3A3')
    thin_white = _ColorFactory('white', '\033[37m', hexValue='#808080')

    success = green
    fail = red
    warning = yellow

    @classmethod
    def color(cls, __val: str, /, *, style='i u r white on red'):
        colors = ['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white']
        thin_colors = [f'thin_{c}' for c in colors]

        def get_color(color: str, background=False):
            background = 10 if background else 0
            try:
                if color.startswith('thin_'):
                    return thin_colors.index(color) + 30 + background
                return colors.index(color) + 90 + background
            except ValueError as e:
                raise ValueError(f'{color!r} not in {colors}')

        style = re.split(r'\s+', style)
        values = set()
        i = 0
        while i < len(style):
            s = style[i]
            if s == 'i':
                values.add(3)
            elif s == 'u':
                values.add(4)
            elif s == 'r':
                values.add(7)
            elif s == 'b' or s == 'bold':
                values.add(1)
            elif s == 'on':
                next_color = style[i + 1]
                values.add(get_color(next_color, True))
                i += 1
            else:
                values.add(get_color(s))
            i += 1

        return _ColorFactory('color', f'\033[{";".join(map(str, values))}m')(__val)


if __name__ == '__main__':
    print(Color.red('hello world'))
    print(Color.green('hello world'))
    print(Color.yellow('hello world'))
    print(Color.blue('hello world'))
    print(Color.cyan('hello world'))
    print(Color.magenta('hello world'))
    print(Color.white('hello world'))
    print(Color.thin_red('hello world'))
    print(Color.thin_green('hello world'))
    print(Color.thin_yellow('hello world'))
    print(Color.thin_blue('hello world'))
    print(Color.thin_cyan('hello world'))
    print(Color.thin_magenta('hello world'))
    print(Color.thin_white('hello world'))
    print('ni hao ' + Color.red('world') + '!')
    print('|'.join(map(str, [Color.red('hello'), Color.green('world')])))
    print(Color.warning('warning: hello world'))
    print('\033[5;34;46mhello world\033[0m')
    print('=' * 50)
    print(Color.color('hello World', style='bold white on black'))
    print(Color.color('hello World', style='on red'))
    print(Color.color('hello World', style='on green'))
    print(Color.color('hello World', style='on yellow'))
    print(Color.color('hello World', style='on blue'))
    print(Color.color('hello World', style='on magenta'))
    print(Color.color('hello World', style='on cyan'))
    print(Color.color('hello World', style='on white'))
    print('=' * 50)
    print(Color.color('Hello World', style='white i u'))
    pass



import re
from typing import Optional

from lljz_tools.color import Color
from lljz_tools.models import Model


class RedisConfig(Model):
    host: [str]
    port: int
    database: int
    username: Optional[str] = None
    password: Optional[str] = None


class RedisClient(Redis):
    def __init__(self, uri: str, ssh_uri: Optional[str] = None):
        self.uri = uri
        self._ssh = None

        self.config = self.parse_uri(uri)  # 解析 URI 得到配置

        # Validate URI
        self.validate_uri()

        if ssh_uri:
            try:
                from lljz_tools.client.utils import setup_ssh_tunnel
            except ImportError:
                raise ImportError(
                    f'package not installed! use "{Color.color('pip install sshtunnel', style='i u yellow')}"')

            self._ssh = setup_ssh_tunnel(self.config.host, self.config.port, ssh_uri)
            self.config.host = '127.0.0.1'
            self.config.port = self._ssh.local_bind_port

        self._redis = self.connect_to_redis()

    @staticmethod
    def parse_uri(uri: str):
        # 解析 URI 并返回配置字典
        pattern = r'redis://(((?P<username>.+):)?(?P<password>.+)@)?(?P<host>.+):(?P<port>\d+)/(?P<database>\d+)?'
        match = re.match(pattern, uri)
        if not match:
            raise ValueError(f'Invalid Redis URI format: {uri}')

        config = match.groupdict()
        config['port'] = int(config['port'])
        config['database'] = int(config['database'])
        return RedisConfig(**config)  # type: ignore

    def validate_uri(self):
        if not self.uri:
            raise ValueError("Redis URI must be specified.")

    def connect_to_redis(self):

        try:
            from redis import Redis

            client = Redis(
                host=self.config.host,
                port=self.config.port,
                username=self.config.username,
                password=self.config.password,
                decode_responses=True,
            )
            return client
        except ImportError:
            raise ImportError(
                f'package not installed! use "{Color.color('pip install redis', style='i u yellow')}"')
        except Exception as e:
            raise ConnectionError(f"Failed to connect to MongoDB: {e}")

    def close(self):
        if self._redis:
            self._redis.close()
        if self._ssh:
            self._ssh.stop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get(self, key):
        return self._redis.get(key)

    def set(self, key, value):
        return self._redis.set(key, value)


if __name__ == '__main__':
    c = RedisClient.parse_uri('redis://123456@192.168.1.220:6379/0')
    print(c)


# coding=utf-8

"""
@fileName       :   columns.py
@data           :   2024/11/15
@author         :   jiangmenggui@hosonsoft.com
"""
import datetime
import re
from typing import Callable, Any, Literal

from dateutil import parser
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation


class ExcelColumn:

    def __init__(
            self, name: str, /, *, required=False, default=None, default_factory: Callable[[], Any] | None = None, **kwargs
    ):
        self.name = name
        self.required = required
        self.default_factory = default_factory or (lambda: default)
        self.kwargs = kwargs

    def validate(self, value):
        pass

    def get_data_validation(self, columnIndex):
        return None


class StringColumn(ExcelColumn):

    def __init__(
            self, name: str, /, *,
            required=False, default=None, default_factory=None, min_length=None, max_length=None, **kwargs
    ):
        super().__init__(
            name, required=required, default=default, default_factory=default_factory,
            min_length=min_length, max_length=max_length, **kwargs
        )

    def validate(self, value):
        if value is None:
            value = self.default_factory()
        if self.required and value is None:
            raise ValueError(f'字段 {self.name} 不能为空')
        if value is None:
            return value
        min_length = self.kwargs.get('min_length')
        max_length = self.kwargs.get('max_length')
        if min_length is not None and len(value) < min_length:
            raise ValueError(f'字段 {self.name} 长度不能小于 {min_length}')
        if max_length is not None and len(value) > max_length:
            raise ValueError(f'字段 {self.name} 长度不能大于 {max_length}')
        return str(value)

    def get_data_validation(self, columnIndex):
        operator: Literal[
            "between", "notBetween", "equal", "notEqual",
            "lessThan", "lessThanOrEqual", "greaterThan",  "greaterThanOrEqual"
        ]
        if self.kwargs.get('max_length') is not None and self.kwargs.get('min_length') is not None:
            operator = 'between'
            formula = dict(formula1=self.kwargs.get('min_length'), formula2=self.kwargs.get('max_length'),)
            errorMessage = f'{self.name}长度必须介于{self.kwargs.get("min_length")}~{self.kwargs.get("max_length")}之间'
        elif self.kwargs.get('min_length') is not None and self.kwargs.get('min_length', 0) > 0:
            operator = 'greaterThanOrEqual'
            formula = {"formula1": self.kwargs.get('min_length')}
            errorMessage = f"{self.name}长度不能少于{self.kwargs.get('min_length')}位"
        elif self.kwargs.get('max_length') is not None and self.kwargs.get('max_length', 0) > 0:
            operator = 'lessThanOrEqual'
            formula = {"formula1": self.kwargs.get('max_length')}
            errorMessage = f"{self.name}长度不能超过{self.kwargs.get('max_length')}位"
        else:
            return None
        dv = DataValidation(
            type='textLength',
            operator=operator,
            **formula,
            showErrorMessage=True,
            showInputMessage=False
        )
        dv.error = errorMessage
        dv.errorTitle = "无效的输入"
        columnName = get_column_letter(columnIndex)
        dv.add(f'{columnName}{2}:{columnName}{10000}')  # 数据验证区域,2-11行
        return dv


class RegexColumn(StringColumn):

    def __init__(
            self, name: str, pattern: str, /, *,
            required=False, default=None, default_factory=None, min_length=None, max_length=None
    ):
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, min_length=min_length,
            max_length=max_length, pattern=pattern
        )

    def validate(self, value):
        value = super().validate(value)
        if value is None:
            return value
        if not re.match(self.kwargs['pattern'], value):
            raise ValueError(f'字段 {self.name} 格式不正确')
        return value


class EmailColumn(RegexColumn):

    def __init__(
            self, name: str, /, *,
            required=False, default=None, default_factory=None, min_length=None, max_length=None
    ):
        super().__init__(
            name, r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$', required=required, default=default,
            default_factory=default_factory, min_length=min_length, max_length=max_length
        )


class IntColumn(ExcelColumn):

    def __init__(
            self, name: str, /, *, required=False, default=None, default_factory=None, max_value: int | None = None,
            min_value: int | None = None, **kwargs
    ):
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, max_value=max_value,
            min_value=min_value, **kwargs
        )

    def validate(self, value):
        if value is None:
            value = self.default_factory()
        if self.required and value is None:
            raise ValueError(f'字段 {self.name} 不能为空')
        if value is None:
            return value
        try:
            value = int(value)
        except ValueError:
            raise ValueError(f'{self.name}必须为数字（实际为{value}）')
        else:
            if self.kwargs.get('max_value') is not None and value > self.kwargs.get('max_value', 0):
                raise ValueError(f'{self.name} 不能大于 {self.kwargs.get("max_value")}（实际为{value}）')
            if self.kwargs.get('min_value') is not None and value < self.kwargs.get('min_value', 0):
                raise ValueError(f'{self.name} 不能小于 {self.kwargs.get("min_value")}（实际为{value}）')
            return value

    def get_data_validation(self, columnIndex):
        operator: Literal[
            "between", "notBetween", "equal", "notEqual",
            "lessThan", "lessThanOrEqual", "greaterThan",  "greaterThanOrEqual"
        ]
        if self.kwargs.get('max_value') is not None and self.kwargs.get('min_value') is not None:
            operator = 'between'
            formula = dict(formula1=self.kwargs.get('min_value'), formula2=self.kwargs.get('max_value'),)
            errorMessage = f'{self.name}只能输入{self.kwargs.get("min_value")}~{self.kwargs.get("max_value")}之间的整数'
        elif self.kwargs.get('min_value') is not None:
            operator = 'greaterThanOrEqual'
            formula = {"formula1": self.kwargs.get('min_value')}
            errorMessage = f"{self.name}不能小于{self.kwargs.get('min_value')}"
        elif self.kwargs.get('max_value') is not None:
            operator = 'lessThanOrEqual'
            formula = {"formula1": self.kwargs.get('max_value')}
            errorMessage = f"{self.name}不能大于{self.kwargs.get('max_value')}"
        else:
            return None
        dv = DataValidation(
            type='whole',
            operator=operator,
            **formula,
            showErrorMessage=True,
            showInputMessage=False
        )
        dv.error = errorMessage
        dv.errorTitle = "无效的输入"
        columnName = get_column_letter(columnIndex)
        dv.add(f'{columnName}{2}:{columnName}{10000}')  # 数据验证区域,2-11行
        return dv


class FloatColumn(ExcelColumn):

    def __init__(
            self, name: str, /, *, required=False, default=None, default_factory=None, max_value: float = None,
            min_value: float | None = None, **kwargs
    ):
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, max_value=max_value,
            min_value=min_value, **kwargs
        )

    def validate(self, value):
        if value is None:
            value = self.default_factory()
        if self.required and value is None:
            raise ValueError(f'字段 {self.name} 不能为空')
        if value is None:
            return value
        try:
            value = float(value)
        except ValueError:
            raise ValueError(f'{self.name}必须为数字（实际为{value}）')
        else:
            if self.kwargs.get('max_value') is not None and value > self.kwargs.get('max_value'):
                raise ValueError(f'{self.name} 不能大于 {self.kwargs.get("max_value")}（实际为{value}）')
            if self.kwargs.get('min_value') is not None and value < self.kwargs.get('min_value'):
                raise ValueError(f'{self.name} 不能小于 {self.kwargs.get("min_value")}（实际为{value}）')
            return value

    def get_data_validation(self, columnIndex):
        operator: Literal[
            "between", "notBetween", "equal", "notEqual",
            "lessThan", "lessThanOrEqual", "greaterThan",  "greaterThanOrEqual"
        ]
        if self.kwargs.get('max_value') is not None and self.kwargs.get('min_value') is not None:
            operator = 'between'
            formula = dict(formula1=self.kwargs.get('min_value'), formula2=self.kwargs.get('max_value'),)
            errorMessage = f'{self.name}只能输入{self.kwargs.get("min_value")}~{self.kwargs.get("max_value")}之间的数字'
        elif self.kwargs.get('min_value') is not None:
            operator = 'greaterThanOrEqual'
            formula = {"formula1": self.kwargs.get('min_value')}
            errorMessage = f"{self.name}不能小于{self.kwargs.get('min_value')}"
        elif self.kwargs.get('max_value') is not None:
            operator = 'lessThanOrEqual'
            formula = {"formula1": self.kwargs.get('max_value')}
            errorMessage = f"{self.name}不能大于{self.kwargs.get('max_value')}"
        else:
            return None
        dv = DataValidation(
            type='decimal',
            operator=operator,
            **formula,
            showErrorMessage=True,
            showInputMessage=False
        )
        dv.error = errorMessage
        dv.errorTitle = "无效的输入"
        columnName = get_column_letter(columnIndex)
        dv.add(f'{columnName}{2}:{columnName}{10000}')  # 数据验证区域,2-11行
        return dv

class DateColumn(ExcelColumn):
    def __init__(
            self, name: str, /, *, required=False, default=None, default_factory=None, max_date: datetime.date = None,
            min_date: datetime.date = None, **kwargs
    ):
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, max_date=max_date,
            min_date=min_date, **kwargs
        )

    def validate(self, value):
        if value is None:
            value = self.default_factory()
        if self.required and value is None:
            raise ValueError(f'字段 {self.name} 不能为空')
        if value is None:
            return value
        try:
            if isinstance(value, datetime.datetime):
                value = value.date()
            elif isinstance(value, datetime.date):
                value = value
            elif isinstance(value, str):
                value = parser.parse(value).date()
            else:
                raise ValueError(f'{self.name}必须为日期格式（实际为{value}）')
        except parser.ParserError:
            raise ValueError(f'{self.name}必须为日期（实际为{value}）')
        else:
            if self.kwargs.get('max_date') is not None and value > self.kwargs.get('max_date'):
                raise ValueError(f'{self.name} 不能大于 {self.kwargs.get("max_date")}（实际为{value}）')
            if self.kwargs.get('min_date') is not None and value < self.kwargs.get('min_date'):
                raise ValueError(f'{self.name} 不能小于 {self.kwargs.get("min_date")}（实际为{value}）')
            return value


class DateTimeColumn(ExcelColumn):
    def __init__(
            self, name: str, /, *, required=False, default=None, default_factory=None, max_date: datetime.date = None,
            min_date: datetime.date = None, **kwargs
    ):
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, max_date=max_date,
            min_date=min_date, **kwargs
        )

    def validate(self, value):
        if value is None:
            value = self.default_factory()
        if self.required and value is None:
            raise ValueError(f'字段 {self.name} 不能为空')
        if value is None:
            return value
        try:
            if isinstance(value, datetime.datetime):
                value = value
            elif isinstance(value, datetime.date):
                value = datetime.datetime.combine(value, datetime.datetime.min.time())
            elif isinstance(value, str):
                value = parser.parse(value)
            else:
                raise ValueError(f'{self.name}必须为日期时间格式（实际为{value}）')
        except parser.ParserError:
            raise ValueError(f'{self.name}必须为日期（实际为{value}）')
        else:
            if self.kwargs.get('max_date') is not None and value > self.kwargs.get('max_date'):
                raise ValueError(f'{self.name} 不能大于 {self.kwargs.get("max_date")}（实际为{value}）')
            if self.kwargs.get('min_date') is not None and value < self.kwargs.get('min_date'):
                raise ValueError(f'{self.name} 不能小于 {self.kwargs.get("min_date")}（实际为{value}）')
            return value


class StringSequenceColumn(StringColumn):
    def __init__(self, name: str, sequence: dict[Any, str] | list[str], /, *, required=False, default=None,
                 default_factory=None,
                 **kwargs):
        if isinstance(sequence, list):
            sequence = {v: v for v in sequence}
        sequence_reverse = {v: k for k, v in sequence.items()}
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, sequence=sequence,
            sequence_reverse=sequence_reverse, **kwargs)

    def validate(self, value):
        value = super().validate(value)
        if value is None:
            return value
        if value not in self.kwargs['sequence_reverse']:
            raise ValueError(f'{self.name}必须在{'/'.join(map(str, self.kwargs["sequence_reverse"].keys()))}中')
        return self.kwargs['sequence_reverse'][value]

    def get_data_validation(self, columnIndex):
        dv = DataValidation(
            type='list',
            formula1=f'"{",".join(self.kwargs['sequence'].values())}"',
            allow_blank=not self.required,
            showErrorMessage=True,
            showInputMessage=False
        )
        dv.error = f"{self.name}只能从【{'、'.join(self.kwargs['sequence'].values())}】中选择"
        dv.errorTitle = "无效的输入"
        dv.prompt = f'请从【{'、'.join(self.kwargs['sequence'].values())}】中选择数据'
        dv.promptTitle = f"选择{self.name}"
        columnName = get_column_letter(columnIndex)
        dv.add(f'{columnName}{2}:{columnName}{10000}')  # 数据验证区域,2-11行
        return dv


class IntSequenceColumn(IntColumn):
    def __init__(self, name: str, sequence: dict[Any, int] | list[int], /, *, required=False, default=None,
                 default_factory=None,
                 **kwargs):
        if isinstance(sequence, list):
            sequence = {v: v for v in sequence}
        sequence_reverse = {v: k for k, v in sequence.items()}
        super().__init__(
            name, required=required, default=default, default_factory=default_factory, sequence=sequence,
            sequence_reverse=sequence_reverse, **kwargs)

    def validate(self, value):
        value = super().validate(value)
        if value is None:
            return value
        if value not in self.kwargs['sequence_reverse']:
            raise ValueError(f'{self.name}必须在{'/'.join(map(str, self.kwargs["sequence_reverse"].keys()))}中')
        return self.kwargs['sequence_reverse'][value]

    def get_data_validation(self, columnIndex):
        dv = DataValidation(
            type='list',
            formula1=f'"{",".join(map(str, self.kwargs['sequence'].values()))}"',
            allow_blank=not self.required,
            showErrorMessage=True,
            showInputMessage=False
        )
        dv.error = f"{self.name}只能从【{'、'.join(self.kwargs['sequence'].values())}】中选择"
        dv.errorTitle = "无效的输入"
        dv.prompt = f'请从【{'、'.join(self.kwargs['sequence'].values())}】中选择数据'
        dv.promptTitle = f"选择{self.name}"
        columnName = get_column_letter(columnIndex)
        dv.add(f'{columnName}{2}:{columnName}{10000}')  # 数据验证区域,2-11行
        return dv


SequenceColumn = StringSequenceColumn

if __name__ == '__main__':
    pass
